#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import sys
from PySide6.QtWidgets import (
    QApplication,
    QMainWindow,
    QTextEdit,
    QFileDialog,
    QToolBar,
    QFontDialog,
    QInputDialog,
    QMessageBox,
    QLabel,
)
from PySide6.QtGui import (
    QTextCharFormat,
    QColor,
    QFont,
    QSyntaxHighlighter,
    QAction,
    QIcon,
)

from PySide6.QtCore import (
    Signal,
)
import re
import revedaEditor.resources.resources
import revedaEditor.backend.dataDefinitions as ddef


class verilogaHighlighter(QSyntaxHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)

        keywordFormat = QTextCharFormat()
        keywordFormat.setForeground(QColor("#0000FF"))
        keywordFormat.setFontWeight(QFont.Bold)
        functionFormat = QTextCharFormat()
        functionFormat.setForeground(QColor("#00FF00"))
        functionFormat.setFontWeight(QFont.Bold)

        self.highlightingRules = [
            (
                r"\b(module|endmodule|input|output|inout|parameter|analog|real|electrical)\b",
                keywordFormat,
            ),
            (
                r"\b(cos|sin|ln|log|min|pow|sinh|sqrt|tan|tanh|exp|cosh|ddt|ddx|idt)\b",
                functionFormat,
            ),
            # Add more highlighting rules as needed
        ]

        self.highlightingComment = QTextCharFormat()
        self.highlightingComment.setForeground(
            QColor("#007F00")
        )  # Green color for comments

        self.commentStart = "//"
        self.multi_comment_start = "/*"
        self.multi_comment_end = "*/"

    def highlightBlock(self, text):
        for pattern, char_format in self.highlightingRules:
            for match in re.finditer(pattern, text):
                self.setFormat(match.start(), match.end() - match.start(), char_format)

        index = text.find(self.commentStart)
        while index >= 0:
            comment_length = len(text) - index
            self.setFormat(index, comment_length, self.highlightingComment)
            index = text.find(self.commentStart, index + comment_length)

        self.setCurrentBlockState(0)
        startIndex = 0
        if self.previousBlockState() != 1:
            startIndex = text.find(self.multi_comment_start)

        while startIndex >= 0:
            endIndex = text.find(self.multi_comment_end, startIndex)
            if endIndex == -1:
                self.setCurrentBlockState(1)
                comment_length = len(text) - startIndex
            else:
                comment_length = endIndex - startIndex + len(self.multi_comment_end)
            self.setFormat(startIndex, comment_length, self.highlightingComment)
            startIndex = text.find(
                self.multi_comment_start, startIndex + comment_length
            )


class xyceHighlighter(QSyntaxHighlighter):
    def __init__(self, parent=None):
        super().__init__(parent)

        keywordFormat = QTextCharFormat()
        keywordFormat.setForeground(QColor("#0000FF"))
        keywordFormat.setFontWeight(QFont.Bold)
        analysisFormat = QTextCharFormat()
        analysisFormat.setForeground(QColor("#00FF0F"))
        analysisFormat.setFontWeight(QFont.Bold)

        self.highlightingRules = [
            (
                r"(\.subckt|\.end|\.ends|\.param|\.model|\.lib|\.sweep)\b",
                keywordFormat,
            ),
            (
                r"\b(\.ac|\.dc|\.tran|\.hb|\.noise|\.op)\b",
                analysisFormat,
            ),
            # Add more highlighting rules as needed
        ]

        self.highlightingComment = QTextCharFormat()
        self.highlightingComment.setForeground(
            QColor("#007F00")
        )  # Green color for comments

        self.comment_start = "*"
        self.multi_comment_start = "/*"
        self.multi_comment_end = "*/"

    def highlightBlock(self, inputText):
        text = inputText.lower()
        for pattern, char_format in self.highlightingRules:
            for match in re.finditer(pattern, text):
                self.setFormat(match.start(), match.end() - match.start(), char_format)

        index = text.find(self.comment_start)
        while index >= 0:
            comment_length = len(text) - index
            self.setFormat(index, comment_length, self.highlightingComment)
            index = text.find(self.comment_start, index + comment_length)

        self.setCurrentBlockState(0)
        startIndex = 0
        if self.previousBlockState() != 1:
            startIndex = text.find(self.multi_comment_start)

        while startIndex >= 0:
            endIndex = text.find(self.multi_comment_end, startIndex)
            if endIndex == -1:
                self.setCurrentBlockState(1)
                comment_length = len(text) - startIndex
            else:
                comment_length = endIndex - startIndex + len(self.multi_comment_end)
            self.setFormat(startIndex, comment_length, self.highlightingComment)
            startIndex = text.find(
                self.multi_comment_start, startIndex + comment_length
            )


class verilogaEditor(QMainWindow):
    closedSignal = Signal(ddef.viewTuple, str)

    def __init__(self, fileName=""):
        super().__init__()
        self._cellViewTuple = ddef.viewTuple("", "", "")

        self.fileName: str = fileName
        self.textEdit = QTextEdit()
        self.setCentralWidget(self.textEdit)

        self.createActions()
        self.createMenus()
        self.createToolbar()

        self.statusLine = self.statusBar()
        self.statusLabel = QLabel()
        self.statusLine.addPermanentWidget(self.statusLabel)

        self.textEdit.cursorPositionChanged.connect(self.updateStatus)

        self.setWindowTitle("Verilog-A Editor")
        self.resize(600, 800)
        self.initEditor()

    def initEditor(self):
        self.highlighter = verilogaHighlighter(self.textEdit.document())
        if self.fileName:
            with open(self.fileName, "r") as file:
                text = file.read()
                self.textEdit.setPlainText(text)

    def createActions(self):
        documentOpenIcon = QIcon(":/icons/document-task.png")
        self.openAction = QAction(documentOpenIcon, "&Open", self)
        self.openAction.setShortcut("Ctrl+O")
        self.openAction.triggered.connect(self.openFile)

        saveIcon = QIcon(":/icons/document.png")
        self.saveAction = QAction(saveIcon, "&Save", self)
        self.saveAction.setShortcut("Ctrl+S")
        self.saveAction.triggered.connect(self.saveFile)

        saveAsIcon = QIcon(":/icons/document--plus.png")
        self.saveAsAction = QAction(saveAsIcon, "Save &As", self)
        self.saveAsAction.triggered.connect(self.saveAsFile)

        quitIcon = QIcon(":/icons/external.png")
        self.quitAction = QAction(quitIcon, "&Quit", self)
        self.quitAction.setShortcut("Ctrl+Q")
        self.quitAction.triggered.connect(self.close)

        fontIcon = QIcon(":icons/ui-label.png")
        self.fontAction = QAction(fontIcon, "&Font", self)
        self.fontAction.setShortcut("Ctrl+F")
        self.fontAction.triggered.connect(self.changeFont)

        cutIcon = QIcon(":/icons/cutter.png")
        self.cutAction = QAction(cutIcon, "&Cut", self)
        self.cutAction.setShortcut("Ctrl+X")
        self.cutAction.triggered.connect(self.textEdit.cut)

        copyIcon = QIcon(":/icons/document-copy.png")
        self.copyAction = QAction(copyIcon, "&Copy", self)
        self.copyAction.setShortcut("Ctrl+C")
        self.copyAction.triggered.connect(self.textEdit.copy)

        pasteIcon = QIcon(":/icons/clipboard-paste.png")
        self.pasteAction = QAction(pasteIcon, "&Paste", self)
        self.pasteAction.setShortcut("Ctrl+V")
        self.pasteAction.triggered.connect(self.textEdit.paste)

        findIcon = QIcon(":/icons/clipboard--pencil.png")
        self.findAction = QAction(findIcon, "&Find", self)
        self.findAction.setShortcut("Ctrl+F")
        self.findAction.triggered.connect(self.findText)

        replaceIcon = QIcon(":/icons/clipboard-search-result.png")
        self.replaceAction = QAction(replaceIcon, "&Replace", self)
        self.replaceAction.setShortcut("Ctrl+H")
        self.replaceAction.triggered.connect(self.replaceText)

    def createMenus(self):
        menubar = self.menuBar()

        fileMenu = menubar.addMenu("&File")
        fileMenu.addAction(self.openAction)
        fileMenu.addAction(self.saveAction)
        fileMenu.addAction(self.saveAsAction)
        fileMenu.addAction(self.quitAction)

        format_menu = menubar.addMenu("&Format")
        format_menu.addAction(self.fontAction)

        editMenu = menubar.addMenu("&Edit")
        editMenu.addAction(self.cutAction)
        editMenu.addAction(self.copyAction)
        editMenu.addAction(self.pasteAction)
        editMenu.addAction(self.findAction)
        editMenu.addAction(self.replaceAction)

    def createToolbar(self):
        toolbar = QToolBar("Toolbar")
        self.addToolBar(toolbar)

        toolbar.addAction(self.openAction)
        toolbar.addAction(self.saveAction)
        toolbar.addAction(self.saveAsAction)
        toolbar.addAction(self.quitAction)
        toolbar.addSeparator()
        toolbar.addAction(self.fontAction)
        toolbar.addSeparator()
        toolbar.addAction(self.cutAction)
        toolbar.addAction(self.copyAction)
        toolbar.addAction(self.pasteAction)
        toolbar.addSeparator()
        toolbar.addAction(self.findAction)
        toolbar.addAction(self.replaceAction)

    def openFile(self):
        (self.fileName, _) = QFileDialog.getOpenFileName(
            self, "Open File", "", "Verilog-A Files (*.va);;All Files (*)"
        )
        if self.fileName:
            with open(self.fileName, "r") as file:
                text = file.read()
                self.textEdit.setPlainText(text)

    def saveFile(self):
        if self.fileName:
            with open(self.fileName, "w") as file:
                text = self.textEdit.toPlainText()
                file.write(text)
        else:
            self.saveAsFile()

    def saveAsFile(self):
        (self.fileName, _) = QFileDialog.getSaveFileName(
            self, "Save File", "", "Verilog-A Files (*.va);;All Files (*)"
        )
        if self.fileName:
            with open(self.fileName, "w") as file:
                text = self.textEdit.toPlainText()
                file.write(text)

    def changeFont(self):
        ok, font = QFontDialog.getFont(self.textEdit.font(), self)
        if ok:
            self.textEdit.setFont(font)

    def findText(self):
        text, ok = QInputDialog.getText(self, "Find", "Enter text to find:")
        if ok:
            document = self.textEdit.document()
            cursor = document.find(text)
            if cursor.isNull():
                QMessageBox.information(self, "Find", "Text not found.")
            else:
                self.textEdit.setTextCursor(cursor)

    def replaceText(self):
        findText, ok1 = QInputDialog.getText(self, "Replace", "Enter text to find:")
        if ok1:
            replace_text, ok2 = QInputDialog.getText(
                self, "Replace", "Enter replacement text:"
            )
            if ok2:
                document = self.textEdit.document()
                cursor = document.find(findText)
                if cursor.isNull():
                    QMessageBox.information(self, "Replace", "Text not found.")
                else:
                    cursor.insertText(replace_text)

    def updateStatus(self):
        cursor = self.textEdit.textCursor()
        line = cursor.blockNumber() + 1
        column = cursor.columnNumber() + 1
        self.statusLabel.setText(f"Line: {line}, Column: {column}")

    def closeEvent(self, event):
        if self.textEdit.toPlainText():
            reply = QMessageBox.question(
                self,
                "Save File",
                "Do you want to save the file?",
                QMessageBox.Yes | QMessageBox.No,
                QMessageBox.No,
            )
            if reply == QMessageBox.Yes:
                self.saveFile()
        self.closedSignal.emit(self.cellViewTuple, self.fileName)
        super().closeEvent(event)

    @property
    def cellViewTuple(self) -> ddef.viewTuple:
        return self._cellViewTuple

    @cellViewTuple.setter
    def cellViewTuple(self, viewTuple):
        if isinstance(viewTuple, ddef.viewTuple):
            self._cellViewTuple = viewTuple


class xyceEditor(verilogaEditor):
    def __init__(self, fileName=""):
        super().__init__(fileName)
        self.initEditor()
        self.setWindowTitle("Xyce/SPICE Editor")

    def initEditor(self):
        self.highlighter = xyceHighlighter(self.textEdit.document())
        if self.fileName:
            with open(self.fileName, "r") as file:
                text = file.read()
                self.textEdit.setPlainText(text)

    def openFile(self):
        (self.fileName, _) = QFileDialog.getOpenFileName(
            self, "Open File", "", "Xyce Files (*.sp);;All Files (*)"
        )
        if self.fileName:
            with open(self.fileName, "r") as file:
                text = file.read()
                self.textEdit.setPlainText(text)

    def saveAsFile(self):
        (self.fileName, _) = QFileDialog.getSaveFileName(
            self, "Save File", "", "Xyce Files (*.sp);;All Files (*)"
        )
        if self.fileName:
            with open(self.fileName, "w") as file:
                text = self.textEdit.toPlainText()
                file.write(text)


def main():
    app = QApplication(sys.argv)
    editor = verilogaEditor()
    editor.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

from PySide6.QtWidgets import QTableView
from PySide6.QtGui import (
    QStandardItemModel,
    QStandardItem,
    QBrush,
    QColor,
    QPixmap,
    QBitmap,
)
from PySide6.QtCore import Signal, Qt, QModelIndex


class layerDataModel(QStandardItemModel):
    def __init__(self, data: list):
        super().__init__()
        self._data = data or []
        self.setColumnCount(5)  # Set the number of columns

        # Set the headers for the columns
        self.setHeaderData(0, Qt.Horizontal, "")
        self.setHeaderData(1, Qt.Horizontal, "Layer")
        self.setHeaderData(2, Qt.Horizontal, "Purpose")
        self.setHeaderData(3, Qt.Horizontal, "V")
        self.setHeaderData(4, Qt.Horizontal, "S")

        for row, layer in enumerate(self._data):
            self.insertRow(row)
            bitmap = QBitmap.fromImage(QPixmap(layer.btexture).scaled(5, 5).toImage())
            brush = QBrush(bitmap)
            brush.setColor(QColor(layer.bcolor))
            item = QStandardItem()
            item.setForeground(QBrush(QColor(255, 255, 255)))
            item.setBackground(brush)
            self.setItem(row, 0, item)
            self.setItem(row, 1, QStandardItem(layer.name))
            self.setItem(row, 2, QStandardItem(layer.purpose))
            item = QStandardItem()
            item.setCheckable(True)
            item.setCheckState(Qt.Checked if layer.selectable else Qt.Unchecked)
            self.setItem(row, 3, item)
            item = QStandardItem()
            item.setCheckable(True)
            item.setCheckState(Qt.Checked if layer.visible else Qt.Unchecked)
            self.setItem(row, 4, item)

    def createData(self, layerlist: list) -> list:
        [
            self._data.append(
                (
                    layer.name,
                    layer.visible,
                    layer.selectable,
                    layer.btexture,
                    layer.bcolor,
                )
            )
            for layer in layerlist
        ]


class layerViewTable(QTableView):
    dataSelected = Signal(str, str)
    layerSelectable = Signal(str, str, bool)
    layerVisible = Signal(str, str, bool)

    def __init__(self, parent=None, model: layerDataModel = None):
        super().__init__(parent)
        self._model = model
        self.setModel(self._model)
        self.resizeColumnsToContents()
        self.setShowGrid(False)
        self.setMaximumWidth(280)
        self.setSelectionBehavior(QTableView.SelectRows)
        self.verticalHeader().setVisible(False)
        selection_model = self.selectionModel()
        selection_model.selectionChanged.connect(self.onSelectionChanged)
        self._model.dataChanged.connect(self.onDataChanged)

    def onSelectionChanged(self, selected, deselected):
        if selected.indexes():
            # Get the first selected index
            layerNameIndex = selected.indexes()[1]
            layerPurposeIndex = selected.indexes()[2]
            # Get the row and column of the selected index
            # Get the data from the model at the selected index
            layerName = self._model.data(layerNameIndex)
            layerPurpose = self._model.data(layerPurposeIndex)
            self.dataSelected.emit(layerName, layerPurpose)

    def onDataChanged(
        self, topLeft: QModelIndex, bottomRight: QModelIndex, roles: list
    ):
        # Check if the changed data involves the check state
        if Qt.CheckStateRole in roles:
            row = topLeft.row()
            column = topLeft.column()
            item = self._model.item(row, column)
            # if item and item.isCheckable():
            if column == 4:
                if item.checkState() == Qt.Checked:
                    self.layerSelectable.emit(
                        self._model.item(row, 1).text(),
                        self._model.item(row, 2).text(),
                        True,
                    )
                else:
                    self.layerSelectable.emit(
                        self._model.item(row, 1).text(),
                        self._model.item(row, 2).text(),
                        False,
                    )
            elif column == 3:
                if item.checkState() == Qt.Checked:
                    self.layerVisible.emit(
                        self._model.item(row, 1).text(),
                        self._model.item(row, 2).text(),
                        True,
                    )
                else:
                    self.layerVisible.emit(
                        self._model.item(row, 1).text(),
                        self._model.item(row, 2).text(),
                        False,
                    )



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#

#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
import json
import logging
import logging.config
import pathlib
import os

from PySide6.QtCore import QThreadPool
from PySide6.QtGui import QAction, QFont, QIcon
from PySide6.QtWidgets import (
    QApplication,
    QDialog,
    QMainWindow,
    QMessageBox,
    QVBoxLayout,
    QWidget,
)

import revedaEditor.backend.hdlBackEnd as hdl
import revedaEditor.backend.importViews as imv
import revedaEditor.gui.editorWindows as edw
import revedaEditor.gui.fileDialogues as fd
import revedaEditor.gui.pythonConsole as pcon
import revedaEditor.gui.stippleEditor as stip
import revedaEditor.gui.helpBrowser as hlp
import revedaEditor.gui.revinit as revinit
import revedaEditor.backend.dataDefinitions as ddef


class mainwContainer(QWidget):
    """
    Definition for the main app window layout.
    """

    def __init__(self, parent):
        super().__init__(parent=parent)
        self.parent = parent
        self.console = pcon.pythonConsole(globals())
        self.init_UI()

    def init_UI(self):
        self.console.setfont(QFont("Fira Mono Regular", 12))
        self.console.writeoutput(
            f"Welcome to Revolution EDA version {revinit.__version__}"
        )
        self.console.writeoutput("Revolution Semiconductor (C) 2024.")
        self.console.writeoutput(
            "Mozilla Public License v2.0 modified with Commons Clause"
        )
        # layout statements, using a grid layout
        gLayout = QVBoxLayout()
        gLayout.setSpacing(10)
        gLayout.addWidget(self.console)
        self.setLayout(gLayout)


#
class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.resize(900, 300)
        self._createActions()
        self._createMenuBar()
        self._createTriggers()
        self.cellViews = [
            "schematic",
            "symbol",
            "layout",
            "veriloga",
            "config",
            "spice",
            "pcell",
        ]
        self.switchViewList = ["schematic", "veriloga", "spice", "symbol"]
        self.stopViewList = ["symbol"]
        self.simulationPath = pathlib.Path.cwd().parent
        self.openViews = dict()
        # create container to position all widgets
        self.centralW = mainwContainer(self)
        self.setCentralWidget(self.centralW)
        self.mainW_statusbar = self.statusBar()
        self.mainW_statusbar.showMessage("Ready")
        self.app = QApplication.instance()
        # print(self.app.reveda_runpathObj)
        # logger name is 'reveda'
        self.logger = logging.getLogger("reveda")
        # library definition file path
        self.runPath = pathlib.Path.cwd()
        # look for library.json file where the script is invoked
        self.libraryPathObj = self.runPath.joinpath("library.json")
        self.libraryDict = self.readLibDefFile(self.libraryPathObj)
        self.libraryBrowser = edw.libraryBrowser(self)
        self.logger_def()
        # revEDAPathObj = Path(__file__)
        # library definition file path
        self.runPath = pathlib.Path.cwd()
        # look for library.json file where the script is invoked
        self.libraryPathObj = self.runPath.joinpath("library.json")
        self.libraryDict = self.readLibDefFile(self.libraryPathObj)
        self.textEditorPath: pathlib.Path = self.runPath
        self.threadPool = QThreadPool.globalInstance()
        self.confFilePath = self.runPath.joinpath("reveda.conf")
        self.loadState()

    def _createMenuBar(self):
        self.mainW_menubar = self.menuBar()
        self.mainW_menubar.setNativeMenuBar(False)
        # Returns QMenu object.
        self.menuFile = self.mainW_menubar.addMenu("&File")
        self.menuTools = self.mainW_menubar.addMenu("&Tools")
        self.importTools = self.menuTools.addMenu("&Import")
        self.menuOptions = self.mainW_menubar.addMenu("&Options")
        self.menuHelp = self.mainW_menubar.addMenu("&Help")
        self.menuFile.addAction(self.exitAction)
        self.menuTools.addAction(self.libraryBrowserAction)
        self.menuTools.addAction(self.createStippleAction)
        self.importTools.addAction(self.importVerilogaAction)
        self.importTools.addAction(self.importSpiceAction)
        self.menuOptions.addAction(self.optionsAction)
        self.menuHelp.addAction(self.helpAction)
        self.menuHelp.addAction(self.aboutAction)
        self.menuOptions.addAction(self.optionsAction)

    def _createActions(self):
        exitIcon = QIcon(":/icons/external.png")
        self.exitAction = QAction(exitIcon, "Exit", self)
        self.exitAction.setShortcut("Ctrl+Q")
        importVerilogaIcon = QIcon(":/icons/document-import.png")
        self.importVerilogaAction = QAction(
            importVerilogaIcon, "Import Verilog-a file..."
        )
        self.importSpiceAction = QAction(
            importVerilogaIcon, "Import Spice file...", self
        )
        openLibIcon = QIcon(":/icons/database--pencil.png")
        self.libraryBrowserAction = QAction(openLibIcon, "Library Browser", self)
        optionsIcon = QIcon(":/icons/resource-monitor.png")
        self.optionsAction = QAction(optionsIcon, "Options...", self)
        self.createStippleAction = QAction("Create Stipple...", self)
        helpIcon = QIcon(":/icons/document-arrow.png")
        self.helpAction = QAction(helpIcon, "Help...", self)
        self.aboutIcon = QIcon(":/icons/information.png")
        self.aboutAction = QAction(self.aboutIcon, "About", self)

    def _createTriggers(self):
        self.exitAction.triggered.connect(self.exitApp)  # type: ignore
        self.libraryBrowserAction.triggered.connect(self.libraryBrowserClick)
        self.importVerilogaAction.triggered.connect(self.importVerilogaClick)
        self.importSpiceAction.triggered.connect(self.importSpiceClick)
        self.optionsAction.triggered.connect(self.optionsClick)
        self.createStippleAction.triggered.connect(self.createStippleClick)
        self.helpAction.triggered.connect(self.helpClick)
        self.aboutAction.triggered.connect(self.aboutClick)

    def readLibDefFile(self, libPath: pathlib.Path):
        libraryDict = dict()
        data = dict()
        if libPath.exists():
            with libPath.open(mode="r") as f:
                data = json.load(f)
            if data.get("libdefs") is not None:
                for key, value in data["libdefs"].items():
                    libraryDict[key] = pathlib.Path(value)
            elif data.get("include") is not None:
                for item in data.get("include"):
                    libraryDict.update(self.readLibDefFile(pathlib.Path(item)))
        return libraryDict

    # open library browser window
    def libraryBrowserClick(self):
        self.libraryBrowser.show()
        self.libraryBrowser.raise_()

    def logger_def(self):
        logging.basicConfig(level=logging.INFO)

        c_handler = logging.StreamHandler(stream=self.centralW.console)
        c_handler.setLevel(logging.INFO)
        c_format = logging.Formatter("%(levelname)s - %(message)s")
        c_handler.setFormatter(c_format)
        f_handler = logging.FileHandler("reveda.log")
        f_handler.setLevel(logging.INFO)
        f_format = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        f_handler.setFormatter(f_format)
        self.logger.addHandler(c_handler)
        self.logger.addHandler(f_handler)

    def closeEvent(self, event):
        reply = QMessageBox.question(
            self,
            "Confirm Exit",
            "Are you sure you want to exit?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )
        if reply == QMessageBox.Yes:
            self.app.closeAllWindows()
        else:
            event.ignore()

    def optionsClick(self):
        dlg = fd.appProperties(self)
        dlg.editorPathEdit.setText(str(self.textEditorPath))
        dlg.simPathEdit.setText(str(self.simulationPath))
        dlg.switchViewsEdit.setText(", ".join(self.switchViewList))
        dlg.stopViewsEdit.setText(", ".join(self.stopViewList))

        if dlg.exec() == QDialog.Accepted:
            self.textEditorPath = pathlib.Path(dlg.editorPathEdit.text())
            self.simulationPath = pathlib.Path(dlg.simPathEdit.text())
            self.switchViewList = [
                switchView.strip()
                for switchView in dlg.switchViewsEdit.text().split(",")
            ]
            self.stopViewList = [
                stopView.strip() for stopView in dlg.stopViewsEdit.text().split(",")
            ]
            if dlg.optionSaveBox.isChecked():
                self.saveState()

    def importVerilogaClick(self):
        """
        Import a Verilog-A view and add it to a design library.

        Args:
            self: The instance of the class.

        Returns:
            None
        """
        self.importVerilogaModule(ddef.viewTuple('', '', ''), '')

    def importVerilogaModule(self, viewT: ddef.viewTuple, filePath: str):
        """

        @param filePath:
        """
        library_model = self.libraryBrowser.libraryModel
        # Open the import dialog
        importDlg = fd.importVerilogaCellDialogue(library_model, self)
        importDlg.vaFileEdit.setText(filePath)
        if viewT.libraryName:
            importDlg.libNamesCB.setCurrentText(viewT.libraryName)
        if viewT.cellName:
            importDlg.cellNamesCB.setCurrentText(viewT.cellName)
        if viewT.viewName:
            importDlg.vaViewName.setText(viewT.viewName)
        else:
            # Set the default view name in the dialog
            importDlg.vaViewName.setText("veriloga")
        # Execute the import dialog and check if it was accepted
        if importDlg.exec() == QDialog.Accepted:
            # Create the Verilog-A object from the file path
            imported_va_obj = hdl.verilogaC(pathlib.Path(importDlg.vaFileEdit.text()))

            # Create the Verilog-A view item tuple
            vaViewItemTuple = imv.createVaView(
                self, importDlg, library_model, imported_va_obj
            )

            # Check if the symbol checkbox is checked
            if importDlg.symbolCheckBox.isChecked():
                # Create the Verilog-A symbol
                imv.createVaSymbol(
                    self,
                    vaViewItemTuple,
                    self.libraryDict,
                    self.libraryBrowser,
                    imported_va_obj,
                )

    def importSpiceClick(self):
        """
        Import a Spice view and add it to a design library.

        Args:
            self: The instance of the class.

        Returns:
            None
        """
        self.importSpiceSubckt("", ddef.viewTuple('', '', ''))

    def importSpiceSubckt(self, viewT:ddef.viewTuple, filePath: str):
        # Get the library model
        library_model = self.libraryBrowser.libraryModel
        # Open the import dialog
        importDlg = fd.importSpiceCellDialogue(library_model, self)
        importDlg.spiceFileEdit.setText(filePath)
        # Set the default view name in the dialog
        if viewT.libraryName:
            importDlg.libNamesCB.setCurrentText(viewT.libraryName)
        if viewT.cellName:
            importDlg.cellNamesCB.setCurrentText(viewT.cellName)
        if viewT.viewName:
            importDlg.spiceViewName.setText(viewT.viewName)
        else:
            importDlg.spiceViewName.setText("spice")
        # Execute the import dialog and check if it was accepted
        if importDlg.exec() == QDialog.Accepted:
            # Create the Verilog-A object from the file path
            importedSpiceObj = hdl.spiceC(pathlib.Path(importDlg.spiceFileEdit.text()))

            # Create the Verilog-A view item tuple
            spiceViewItemTuple = imv.createSpiceView(
                self, importDlg, library_model, importedSpiceObj
            )

            # Check if the symbol checkbox is checked
            if importDlg.symbolCheckBox.isChecked():
                # Create the spice symbol
                imv.createSpiceSymbol(
                    self,
                    spiceViewItemTuple,
                    self.libraryDict,
                    self.libraryBrowser,
                    importedSpiceObj,
                )

    def createStippleClick(self):
        stippleWindow = stip.stippleEditor(self)
        stippleWindow.show()

    def helpClick(self):
        helpBrowser = hlp.helpBrowser(self)
        helpBrowser.show()

    def aboutClick(self):
        abtDlg = hlp.aboutDialog(self)
        abtDlg.show()

    def loadState(self):
        """
        Load the state of the object from a configuration file.

        This function reads the contents of the configuration file and updates the
         state of the object based on the values found in the file.
        It checks if the configuration file exists and then opens it for reading.
        If the file exists, it loads the contents of the file as a
        JSON object and assigns the values to the corresponding attributes of the
        object. The attributes updated include `textEditorPath`,
        `simulationPath`, `switchViewList`, and `stopViewList`. If the `switchViewList`
        or `stopViewList` in the configuration file is not
        empty, it updates the corresponding attributes with the values from the file.
        """

        if self.confFilePath.exists():
            self.logger.info(f"Configuration file: {self.confFilePath} exists")
            with self.confFilePath.open(mode="r") as f:
                items = json.load(f)
            self.textEditorPath = pathlib.Path(items.get("textEditorPath"))
            self.simulationPath = pathlib.Path(items.get("simulationPath"))
            if items.get("switchViewList")[0] != "":
                self.switchViewList = items.get("switchViewList")
            if items.get("stopViewList")[0] != "":
                self.stopViewList = items.get("stopViewList")

    def saveState(self):
        items = {
            "textEditorPath": str(self.textEditorPath),
            "simulationPath": str(self.simulationPath),
            "switchViewList": self.switchViewList,
            "stopViewList": self.stopViewList,
        }
        with self.confFilePath.open(mode="w", encoding="utf") as f:
            json.dump(items, f, indent=4)

    def exitApp(self):
        self.app.closeAllWindows()



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#   #
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#   #
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#   #
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)

from typing import NamedTuple, Union
from dataclasses import dataclass

from PySide6.QtCore import Qt, QPoint, QPointF
from PySide6.QtGui import QColor


@dataclass
class edLayer:
    name: str = ""  # edLayer name
    purpose: str = "drawing"  # edLayer purpose
    pcolor: QColor = Qt.black  # pen colour
    pwidth: int = 1  # pen width
    pstyle: Qt.PenStyle = Qt.SolidLine  # pen style
    bcolor: QColor = Qt.transparent  # brush colour
    bstyle: Qt.BrushStyle = Qt.SolidPattern  # brush texture
    z: int = 1  # z-index
    selectable: bool = True  # selectable
    visible: bool = True  # visible
    gdsLayer: int = 0  # gds edLayer
    datatype: int = 0  # gds datatype


@dataclass
class layLayer:
    name: str = ""  # edLayer name
    purpose: str = "drawing"  # edLayer purpose
    pcolor: QColor = Qt.black  # pen colour
    pwidth: int = 1  # pen width
    pstyle: Qt.PenStyle = Qt.SolidLine  # pen style
    bcolor: QColor = Qt.transparent  # brush colour
    btexture: str = ""  # brush texture
    z: int = 1  # z-index
    selectable: bool = True  # selectable
    visible: bool = True  # visible
    gdsLayer: int = 0  # gds edLayer
    datatype: int = 0  # gds datatype


@dataclass
class editModes:
    selectItem: bool
    deleteItem: bool
    moveItem: bool
    copyItem: bool
    rotateItem: bool
    changeOrigin: bool
    panView: bool
    stretchItem: bool

    def setMode(self, attribute):
        for key in self.__dict__.keys():
            self.__dict__[key] = False
        self.__dict__[attribute] = True


@dataclass
class symbolModes(editModes):
    drawPin: bool
    drawArc: bool
    drawRect: bool
    drawLine: bool
    addLabel: bool
    drawCircle: bool
    drawPolygon: bool


@dataclass
class schematicModes(editModes):
    drawPin: bool
    drawWire: bool
    drawText: bool
    addInstance: bool


@dataclass
class layoutModes(editModes):
    drawPath: bool
    drawPin: bool
    drawArc: bool
    drawPolygon: bool
    addLabel: bool
    addVia: bool
    drawRect: bool
    drawLine: bool
    drawCircle: bool
    drawRuler: bool
    addInstance: bool


# library editor related named tuples
class viewTuple(NamedTuple):
    libraryName: str
    cellName: str
    viewName: str


class cellTuple(NamedTuple):
    libraryName: str
    cellName: str


class viewItemTuple(NamedTuple):
    libraryItem: object
    cellItem: object
    viewItem: object


# layout related tuples
class layoutPathTuple(NamedTuple):
    layer: layLayer
    name: str
    mode: str
    width: float
    startExtend: bool
    endExtend: bool


class layoutPinTuple(NamedTuple):
    pinName: str
    pinDir: str
    pinType: str
    pinLayer: layLayer


class layoutLabelTuple(NamedTuple):
    labelText: str
    fontFamily: str
    fontStyle: str
    fontHeight: float
    labelAlign: str
    labelOrient: str
    labelLayer: str


class rulerTuple(NamedTuple):
    point: Union[QPoint, QPointF]
    line: tuple
    text: str


# # pdk related classes and namedtuples
# # this tuple defines the minimum dimensions of a via
# # This can be extended to define the maximum dimensions


class viaDefTuple(NamedTuple):
    name: str
    layer: layLayer
    type: str
    minWidth: float
    maxWidth: float
    minHeight: float
    maxHeight: float
    minSpacing: float
    maxSpacing: float


class singleViaTuple(NamedTuple):
    viaDefTuple: viaDefTuple
    width: float
    height: float


class arrayViaTuple(NamedTuple):
    singleViaTuple: singleViaTuple
    spacing: float
    xnum: int
    ynum: int


# rectangle coordinates tuple


class rectCoords(NamedTuple):
    left: float
    top: float
    w: float
    h: float




#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

from quantiphy import Quantity

class baseInst():
    def __init__(self, labels_dict: dict):
        self._labelsDict = labels_dict

class res(baseInst):
    def __init__(self,labels_dict:dict):
        super().__init__(labels_dict)

    def doubleR(self):
        Rvalue = self._labelsDict.get('R').labelValue
        if Rvalue.isalnum():
            return str(2*Quantity(Rvalue))
        return '?'

class nmos(baseInst):
    def __init__(self,labels_dict:dict):
        super().__init__(labels_dict)
        self.w = Quantity(self._labelsDict['@w'].labelValue)
        self.l = Quantity(self._labelsDict['@l'].labelValue)
        self.nf= Quantity(self._labelsDict['@nf'].labelValue)
        self.sd1p8v = 0.28
        self.sa1p8v = sb1p8v = 0.265
        self.sourceDiffs = lambda nf: int(int(nf) / 2 + 1)

    def asparm(self):
        return self.sourceDiffs(self.nf)*(self.w/self.nf)*self.sd1p8v



#    “Commons Clause” License Condition v1.0
#
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

from PySide6.QtCore import QPoint, Qt
from PySide6.QtGui import QBrush, QFont, QColor
from PySide6.QtWidgets import QGraphicsSimpleTextItem, QGraphicsItem
import pdk.symLayers as symlyr
import pdk.callbacks as cb
import itertools as itt
from typing import Union, Optional, NamedTuple
from quantiphy import Quantity


class symbolLabel(QGraphicsSimpleTextItem):
    """
    label: text class definition for symbol drawing.
    labelText is what is shown on the symbol in a schematic
    """

    labelAlignments = ["Left", "Center", "Right"]
    labelOrients = ["R0", "R90", "R180", "R270", "MX", "MX90", "MY", "MY90"]
    labelUses = ["Normal", "Instance", "Pin", "Device", "Annotation"]
    labelTypes = ["Normal", "NLPLabel", "PyLabel"]
    predefinedLabels = [
        "[@libName]",
        "[@cellName]",
        "[@viewName]",
        "[@instName]",
        "[@modelName]",
        "[@elementNum]",
    ]

    def __init__(
        self,
        start: QPoint,
        labelDefinition: str,
        labelType: str,
        labelHeight: str,
        labelAlign: str,
        labelOrient: str,
        labelUse: str,
    ):
        super().__init__("")
        self._start = start  # top left corner
        self._labelDefinition = labelDefinition  # label definition is what is
        # entered in the symbol editor
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self.setFlag(QGraphicsItem.ItemIsFocusable, True)
        self.setAcceptHoverEvents(True)
        self._labelName = ""  # label Name
        self._labelValue = ""  # label value
        self._labelText = ""  # Displayed label
        self._labelHeight = labelHeight
        self._labelAlign = labelAlign
        self._labelOrient = labelOrient
        self._labelUse = labelUse
        self._labelType = labelType
        self._labelFont = QFont("Arial")
        self._labelFont.setPointSize(int(float(self._labelHeight)))
        self._labelFont.setKerning(False)
        self._labelVisible: bool = False

        self._angle: float = 0.0  # rotation angle
        self.setBrush(symlyr.labelBrush)
        self.setPos(self._start)

        match self._labelOrient:
            case "R0":
                self.setRotation(0)
            case "R90":
                self.setRotation(90)
            case "R180":
                self.setRotation(180)
            case "R270":
                self.setRotation(270)
            case _:
                self.setRotation(0)

    def __repr__(self):
        return (
            f"symbolLabel({self._start},{self._labelDefinition},"
            f" {self._labelType}, {self._labelHeight}, {self._labelAlign}, {self._labelOrient},"
            f" {self._labelUse})"
        )

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemSelectedHasChanged:
            if value:
                self.setBrush(symlyr.selectedLabelBrush)
            else:
                self.setBrush(symlyr.labelBrush)
        return super().itemChange(change, value)

    def contextMenuEvent(self, event):
        self.scene().itemContextMenu.exec_(event.screenPos())

    @property
    def start(self) -> QPoint:
        return self._start

    @start.setter
    def start(self, start: QPoint):
        self._start = start

    @property
    def labelName(self):
        return self._labelName

    @labelName.setter
    def labelName(self, labelName: str):
        self._labelName = labelName

    @property
    def labelDefinition(self):
        return self._labelDefinition

    @labelDefinition.setter
    def labelDefinition(self, labelDefinition: str):
        if isinstance(labelDefinition, str):
            self._labelDefinition = labelDefinition.strip()
            self.labelDefs()

    @property
    def labelValue(self):
        return self._labelValue

    @labelValue.setter
    def labelValue(self, labelValue):
        self._labelValue = labelValue
        # if label value is set.
        self.labelDefs()

    @property
    def labelText(self):
        return self._labelText

    @labelText.setter
    def labelText(self, labelText):
        self._labelText = labelText
        self.labelDefs()

    @property
    def labelType(self):
        return self._labelType

    @labelType.setter
    def labelType(self, labelType):
        if labelType in self.labelTypes:
            self._labelType = labelType
            self.labelDefs()
        elif self.scene():
            self.scene().logger.error("Invalid label type")

    @property
    def labelAlign(self):
        return self._labelAlign

    @labelAlign.setter
    def labelAlign(self, labelAlignment):
        if labelAlignment in self.labelAlignments:
            self._labelAlign = labelAlignment
        elif self.scene():
            self.scene().logger.error("Invalid label alignment")

    @property
    def labelHeight(self):
        return self._labelHeight

    @labelHeight.setter
    def labelHeight(self, labelHeight):
        self.prepareGeometryChange()
        self._labelHeight = labelHeight
        self._labelFont.setPointSize(int(float(labelHeight)))
        self.setFont(self._labelFont)

    @property
    def labelOrient(self):
        return self._labelOrient

    @labelOrient.setter
    def labelOrient(self, labelOrient):
        if labelOrient in self.labelOrients:
            self._labelOrient = labelOrient
        else:
            self.scene().logger.error("Invalid label orientation")

    @property
    def labelUse(self):
        return self._labelUse

    @labelUse.setter
    def labelUse(self, labelUse):
        if labelUse in self.labelUses:
            self._labelUse = labelUse
        elif self.scene():
            self.scene().logger.error("Invalid label use")

    @property
    def labelFont(self):
        return self._labelFont

    @labelFont.setter
    def labelFont(self, labelFont: QFont):
        self._labelFont = labelFont

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value):
        self._angle = value
        self.prepareGeometryChange()
        self.setRotation(value)

    @property
    def labelVisible(self) -> bool:
        return self._labelVisible

    @labelVisible.setter
    def labelVisible(self, value: bool):
        assert isinstance(value, bool)
        if value:
            self.setOpacity(1)
            self._labelVisible = True
        else:
            self.setOpacity(0.001)
            self._labelVisible = False

    def moveBy(self, delta: QPoint):
        self._start += delta

    def labelDefs(self):
        """
        This method creates label name, value, and text from a label definition.
        It should be called when a label is defined or redefined.
        """
        self.prepareGeometryChange()

        if self._labelType == symbolLabel.labelTypes[0]:  # normal label
            # Set label name, value, and text to label definition
            self._labelName = f"@{self._labelDefinition}"
            self._labelValue = self._labelDefinition
            self._labelText = self._labelDefinition
        elif self._labelType == symbolLabel.labelTypes[1]:  # NLPLabel
            self.createNLPLabel()
        elif self._labelType == symbolLabel.labelTypes[2]:  # pyLabel
            # self._labelText = f'{self._labelDefinition}'
            self.createPyLabel()
        self.setText(self._labelText)

    def createNLPLabel(self):
        try:
            if self._labelDefinition.strip().startswith("[@"):
                # Find the end of the expression
                end_index = self._labelDefinition.find("]")
                expression = self._labelDefinition[1:end_index]
                parts = expression.split(":")
                self._labelName = parts[0].strip()
                if self.parentItem() is None:  # symbol editor
                    self._labelText = self._labelDefinition
                    self._labelValue = ""
                else:
                    if self._labelDefinition in symbolLabel.predefinedLabels:
                        match self._labelDefinition:
                            case "[@cellName]":
                                # Set label name to "cellName" and value and text to parent item's cell name
                                self._labelValue = self.parentItem().cellName
                                self._labelText = self._labelValue
                            case "[@instName]":
                                # Set label name to "instName" and value and text to parent item's counter with prefix "I"
                                self._labelValue = f"I{self.parentItem().counter}"
                                self._labelText = self._labelValue
                            case "[@libName]":
                                # Set label name to "libName" and value and text to parent item's library name
                                self._labelValue = self.parentItem().libraryName
                                self._labelText = self._labelValue
                            case "[@viewName]":
                                # Set label name to "viewName" and value and text to parent item's view name
                                self._labelValue = self.parentItem().viewName
                                self._labelText = self._labelValue
                            case "[@modelName]":
                                # Set label name to "modelName" and value and text to parent item's "modelName" attribute
                                self._labelValue = self.parentItem().attr.get(
                                    "modelName", ""
                                )
                                self._labelText = self._labelValue
                            case "[@elementNum]":
                                # Set label name to "elementNum" and value and text to parent item's counter
                                self._labelValue = f"{self.parentItem().counter}"
                                self._labelText = self._labelValue
                    else:
                        if len(parts) > 1:
                            formatString = parts[1]
                            defaultValue = parts[2] if len(parts) > 2 else ""
                        else:
                            formatString = ""
                            defaultValue = ""

                        if formatString:
                            prefix, suffix = formatString.split("%")
                            if self._labelValue:
                                self._labelText = f"{prefix}{self._labelValue}{suffix}"
                            elif defaultValue:
                                self._labelValue = defaultValue.replace(prefix, "")
                                self._labelText = defaultValue
                            else:
                                self._labelValue = "?"
                                self._labelText = f"{prefix}{self._labelValue}{suffix}"
        except Exception as e:
            self.scene().logger.error(
                f"Error parsing label definition: {self._labelDefinition}, {e}"
            )

    # def createPyLabel(self):
    #     try:
    #         labelFields = self._labelDefinition.strip().split("=")
    #         self._labelName = labelFields[0].strip()
    #         labelFunction = labelFields[1].strip()
    #         if self.parentItem() and hasattr(self.parentItem(), "cellName"):
    #             expression = f"cb.{self.parentItem().cellName}(self.parentItem().labels).{labelFunction}"
    #             self._labelValue = Quantity(eval(expression)).render(prec=3)
    #             self._labelText = f"{self._labelName}={self._labelValue}"
    #         else:
    #             self._labelText = f"{self._labelName} = {labelFunction}"
    #     except Exception as e:
    #         if self.scene():
    #             self.scene().logger.error(f"PyLabel Error:{e}")

    def createPyLabel(self):
        """
        Create a PyLabel using the label definition and parent item information.
        """
        try:
            # Split the label definition into name and function
            labelName, labelFunction = map(str.strip, self._labelDefinition.split("="))

            # Check if parent item exists and has 'cellName' attribute
            if self.parentItem() and hasattr(self.parentItem(), "cellName"):
                # Construct the expression to evaluate
                expression = f"cb.{self.parentItem().cellName}(self.parentItem().labels).{labelFunction}"
                # Evaluate the expression and render the result
                self._labelValue = Quantity(eval(expression)).render(prec=3)
                # Set the label text with the name and value
                self._labelText = f"{labelName}={self._labelValue}"
            else:
                # Set the label text with the name and function
                self._labelText = f"{labelName} = {labelFunction}"
            self._labelName = f"@{labelName}"
        except Exception as e:
            # Log the error if scene exists
            if self.scene():
                self.scene().logger.error(f"PyLabel Error:{e}")



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#
# import pdk.layoutLayers as laylyr
from PySide6.QtCore import (
    QPoint,
)

import pdk.layoutLayers as laylyr
import pdk.process as fabproc
import revedaEditor.common.layoutShapes as lshp

class nmos(lshp.layoutPcell):
    cut = int(0.17 * fabproc.dbu)
    poly_to_cut = int(0.055 * fabproc.dbu)
    diff_ovlp_cut = int(0.06 * fabproc.dbu)
    poly_ovlp_diff = int(0.13 * fabproc.dbu)
    nsdm_ovlp_diff = int(0.12 * fabproc.dbu)
    li_ovlp_cut = int(0.06 * fabproc.dbu)
    sa = poly_to_cut + cut + diff_ovlp_cut
    sd = 2 * (max(poly_to_cut, diff_ovlp_cut)) + cut
    # when initialized it has no shapes. 
    def __init__(
            self,
            width: str = 4.0,
            length: str = 0.13,
            nf: str = 1,
    ):
        self._shapes = []
        # define the device parameters here but set them to zero
        self._deviceWidth = float(width) # device width
        self._drawnWidth: int = int(fabproc.dbu * self._deviceWidth) # width in grid points
        self._deviceLength = float(length) # gate length
        self._drawnLength: int = int(fabproc.dbu * self._deviceLength)
        self._nf = int(float(nf)) # number of fingers.
        self._widthPerFinger = int(self._drawnWidth / self._nf)
        super().__init__(self._shapes)
    #

    def __call__(self, width:float, length:float, nf:int):
        '''
        When pcell instance is called, it removes all the shapes and recreates them and adds them as child items to pcell.
        '''
        self._deviceWidth = float(width) # total gate width
        self._drawnWidth = int(self._deviceWidth * fabproc.dbu) # drawn gate width in grid points
        self._deviceLength = float(length) # gate length
        self._drawnLength = int(self._deviceLength * fabproc.dbu) # drawn gate length in grid points
        self._nf = int(float(nf)) # number of fingers
        self._widthPerFinger = self._drawnWidth / self._nf
        self.shapes = self.createGeometry()

    def createGeometry(self) -> list[lshp.layoutShape]:
        activeRect = lshp.layoutRect(
            QPoint(0, 0),
            QPoint(
                self._widthPerFinger,
                int(self._nf * self._drawnLength + 2 * nmos.sa + (self._nf - 1) * nmos.sd),
            ),
            laylyr.odLayer_drw,
        )
        polyFingers = [lshp.layoutRect(
            QPoint(-nmos.poly_ovlp_diff,
            nmos.sa + finger * (self._drawnLength + nmos.sd)),
            QPoint(self._widthPerFinger + nmos.poly_ovlp_diff,
            nmos.sa + finger * (self._drawnLength + nmos.sd) + self._drawnLength), laylyr.poLayer_drw,
        ) for finger in range(self._nf)]
        # contacts = [lshp.layoutRect(
            
        # )]
        return [activeRect, *polyFingers]

    @property
    def width(self):
        return self._deviceWidth

    @width.setter
    def width(self, value: float):
        self._deviceWidth = value

    @property
    def length(self):
        return self._deviceLength

    @length.setter
    def length(self, value: float):
        self._deviceLength = value

    @property
    def nf(self):
        return self._nf

    @nf.setter
    def nf(self, value: int):
        self._nf = value

class pmos(lshp.layoutPcell):
    pass



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import math
from functools import cached_property

from PySide6.QtCore import (
    QPoint,
    QRect,
    QRectF,
    Qt,
    QLine,
    QLineF,
)
from PySide6.QtGui import (
    QBrush,
    QFont,
    QFontMetrics,
    QPainterPath,
    QTextOption,
    QFontDatabase,
    QTransform,
    QPolygonF,
)
from PySide6.QtWidgets import (
    QGraphicsItem,
    QGraphicsSceneMouseEvent,
    QGraphicsSceneHoverEvent,
)
from quantiphy import Quantity
from typing import (
    Union,
    NamedTuple,
)
import pdk.schLayers as schlyr
import pdk.symLayers as symlyr

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.common.net as net
from revedaEditor.common.labels import symbolLabel


class symbolShape(QGraphicsItem):
    def __init__(self) -> None:
        super().__init__()
        self.setFlag(QGraphicsItem.ItemIsMovable, False)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self.setFlag(QGraphicsItem.ItemIsFocusable, True)
        self.setAcceptHoverEvents(True)
        self._angle: float = 0.0  # rotation angle
        self._stretch: bool = False
        self._pen = symlyr.defaultPen
        self._draft: bool = False
        self._brush: QBrush = schlyr.draftBrush

    def __repr__(self):
        return "symbolShape()"

    @property
    def pen(self):
        return self._pen

    @property
    def brush(self):
        return self._brush

    @brush.setter
    def brush(self, value: QBrush):
        if isinstance(value, QBrush):
            self._brush = value

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value):
        self._angle = value
        self.prepareGeometryChange()
        self.setRotation(value)

    @property
    def stretch(self):
        return self._stretch

    @stretch.setter
    def stretch(self, value: bool):
        self._stretch = value

    @property
    def draft(self) -> bool:
        return self._draft

    @draft.setter
    def draft(self, value: bool):
        assert isinstance(value, bool)
        self._draft = value
        # all the child items and their children should be also draft.
        for item in self.childItems():
            item.draft = True

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        super().mousePressEvent(event)

    def sceneEvent(self, event):
        """
        Do not propagate event if shape needs to keep still.
        """
        if self.scene() and (
            self.scene().editModes.changeOrigin or self.scene().drawMode
        ):
            return False
        else:
            super().sceneEvent(event)
            return True

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)  # self.setSelected(False)
        self.setFlag(QGraphicsItem.ItemIsMovable, False)

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        self.setCursor(Qt.ArrowCursor)
        self.setOpacity(0.75)
        self.setFocus()
        super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        super().hoverLeaveEvent(event)
        self.setCursor(Qt.CrossCursor)
        self.setOpacity(1)
        self.clearFocus()

    def contextMenuEvent(self, event):
        self.scene().itemContextMenu.exec_(event.screenPos())


class symbolRectangle(symbolShape):
    """
        rect: QRect defined by top left corner and bottom right corner. QRect(Point1,Point2)
    f"""

    sides = ["Left", "Right", "Top", "Bottom"]

    def __init__(self, start: QPoint, end: QPoint) -> None:
        super().__init__()
        self._rect = QRectF(start, end).normalized()
        self._start = self._rect.topLeft()
        self._end = self._rect.bottomRight()
        self._stretchSide = None
        self._pen = symlyr.symbolPen

    def boundingRect(self):
        return self._rect.normalized().adjusted(-2, -2, 2, 2)

    def paint(self, painter, option, widget):
        if self.draft:
            painter.setPen(symlyr.draftPen)
            self.setZValue(symlyr.draftLayer.z)
        elif self.isSelected():
            painter.setPen(symlyr.selectedSymbolPen)
            self.setZValue(symlyr.symbolLayer.z)
            if self.stretch:
                painter.setPen(symlyr.stretchSymbolPen)
                self.setZValue(symlyr.stretchSymbolLayer.z)
                if self._stretchSide == symbolRectangle.sides[0]:
                    painter.drawLine(self.rect.topLeft(), self.rect.bottomLeft())
                elif self._stretchSide == symbolRectangle.sides[1]:
                    painter.drawLine(self.rect.topRight(), self.rect.bottomRight())
                elif self._stretchSide == symbolRectangle.sides[2]:
                    painter.drawLine(self.rect.topLeft(), self.rect.topRight())
                elif self._stretchSide == symbolRectangle.sides[3]:
                    painter.drawLine(self.rect.bottomLeft(), self.rect.bottomRight())
        else:
            painter.setPen(symlyr.symbolPen)
            self.setZValue(symlyr.symbolLayer.z)
        painter.drawRect(self._rect)

    def __repr__(self):
        return f"symbolRectangle({self._start},{self._end})"

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, rect: QRect):
        self.prepareGeometryChange()
        self._rect = rect

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start: QPoint):
        self.prepareGeometryChange()
        self._rect = QRectF(start, self.end).normalized()
        self._start = start

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end: QPoint):
        self.prepareGeometryChange()
        self._rect = QRectF(self.start, end).normalized()
        self._end = end

    @property
    def centre(self):
        return QPoint(
            int(self._rect.x() + self._rect.width() / 2),
            int(self._rect.y() + self._rect.height() / 2),
        )

    @property
    def height(self):
        return self._rect.height()

    @height.setter
    def height(self, height: int):
        self.prepareGeometryChange()
        self._rect.setHeight(height)

    @property
    def width(self):
        return self._rect.width()

    @width.setter
    def width(self, width):
        self.prepareGeometryChange()
        self._rect.setWidth(width)

    @property
    def left(self):
        return self._rect.left()

    @left.setter
    def left(self, left: int):
        self._rect.setLeft(left)

    @property
    def right(self):
        return self._rect.right()

    @right.setter
    def right(self, right: int):
        self.prepareGeometryChange()
        self._rect.setRight(right)

    @property
    def top(self):
        return self._rect.top()

    @top.setter
    def top(self, top: int):
        self.prepareGeometryChange()
        self._rect.setTop(top)

    @property
    def bottom(self):
        return self._rect.bottom()

    @bottom.setter
    def bottom(self, bottom: int):
        self.prepareGeometryChange()
        self._rect.setBottom(bottom)

    @property
    def origin(self):
        return self._rect.bottomLeft()

    @property
    def stretchSide(self):
        return self._stretchSide

    @stretchSide.setter
    def stretchSide(self, value: str):
        self.prepareGeometryChange()
        self._stretchSide = value

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        eventPos = event.pos().toPoint()
        if self._stretch:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            if eventPos.x() == self._rect.left():
                if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                    self.setCursor(Qt.SizeHorCursor)
                    self._stretchSide = symbolRectangle.sides[0]
            elif eventPos.x() == self._rect.right():
                if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                    self.setCursor(Qt.SizeHorCursor)
                    self._stretchSide = symbolRectangle.sides[1]
            elif eventPos.y() == self._rect.top():
                if self._rect.left() <= eventPos.x() <= self._rect.right():
                    self.setCursor(Qt.SizeVerCursor)
                    self._stretchSide = symbolRectangle.sides[2]
            elif eventPos.y() == self._rect.bottom():
                if self._rect.left() <= eventPos.x() <= self._rect.right():
                    self.setCursor(Qt.SizeVerCursor)
                    self._stretchSide = symbolRectangle.sides[3]

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        eventPos = event.pos().toPoint()
        if self.stretch:
            self.prepareGeometryChange()
            if self.stretchSide == symbolRectangle.sides[0]:
                self.setCursor(Qt.SizeHorCursor)
                self.rect.setLeft(eventPos.x())
            elif self.stretchSide == symbolRectangle.sides[1]:
                self.setCursor(Qt.SizeHorCursor)
                self.rect.setRight(eventPos.x() - int(self._pen.width() / 2))
            elif self.stretchSide == symbolRectangle.sides[2]:
                self.setCursor(Qt.SizeVerCursor)
                self.rect.setTop(eventPos.y())
            elif self.stretchSide == symbolRectangle.sides[3]:
                self.setCursor(Qt.SizeVerCursor)
                self.rect.setBottom(eventPos.y() - int(self._pen.width() / 2))
            self.update()
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        super().mouseReleaseEvent(event)
        # self.start = self.rect.topLeft()
        # self.end = self.rect.bottomRight()
        if self.stretch:
            self._stretch = False
            self._stretchSide = None
            self.setCursor(Qt.ArrowCursor)


class symbolCircle(symbolShape):
    def __init__(self, centre: QPoint, end: QPoint):
        super().__init__()
        xlen = abs(end.x() - centre.x())
        ylen = abs(end.y() - centre.y())
        self._radius = int(math.sqrt(xlen**2 + ylen**2))
        self._centre = centre
        self._topLeft = self._centre - QPoint(self._radius, self._radius)
        self._rightBottom = self._centre + QPoint(self._radius, self._radius)
        self._end = self._centre + QPoint(self._radius, 0)  # along x-axis
        self._stretch = False
        self._startStretch = False

    def paint(self, painter, option, widget) -> None:
        if self.isSelected():
            painter.setPen(symlyr.selectedSymbolPen)
            self.setZValue(symlyr.selectedSymbolLayer.z)
            painter.drawEllipse(self._centre, 1, 1)
            if self._stretch:
                painter.setPen(symlyr.stretchSymbolPen)
                self.setZValue(symlyr.stretchSymbolLayer.z)
        else:
            painter.setPen(symlyr.symbolPen)
            self.setZValue(symlyr.symbolLayer.z)
        painter.drawEllipse(self._centre, self._radius, self._radius)

    def __repr__(self):
        return f"symbolCircle({self._centre},{self._end})"

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, radius: int):
        self.prepareGeometryChange()
        self._radius = radius
        self._end = self._centre + QPoint(self._radius, 0)
        self._topLeft = self._centre - QPoint(self._radius, self._radius)
        self._rightBottom = self._centre + QPoint(self._radius, self._radius)

    @property
    def centre(self):
        return self._centre

    @centre.setter
    def centre(self, value: QPoint):
        self.prepareGeometryChange()
        if isinstance(value, QPoint):
            self._centre = value

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value: QPoint):
        if isinstance(value, QPoint):
            self.prepareGeometryChange()
            self._end = value

    @property
    def rightBottom(self):
        return self._rightBottom

    @rightBottom.setter
    def rightBottom(self, value: QPoint):
        if isinstance(value, QPoint):
            self._rightBottom = value

    @property
    def topLeft(self):
        return self._topLeft

    @topLeft.setter
    def topLeft(self, value: QPoint):
        if isinstance(value, QPoint):
            self._topLeft = value

    def boundingRect(self):
        return (
            QRectF(self._topLeft, self._rightBottom).normalized().adjusted(-2, -2, 2, 2)
        )

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self.isSelected() and self._stretch:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            eventPos = event.pos().toPoint()
            distance = math.sqrt(
                (eventPos.x() - self._centre.x()) ** 2
                + (eventPos.y() - self._centre.y()) ** 2
            )
            if distance == self._radius:
                self._startStretch = True
                self.setCursor(Qt.DragMoveCursor)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseMoveEvent(event)
        if self._startStretch:
            eventPos = event.pos().toPoint()
            distance = math.sqrt(
                (eventPos.x() - self._centre.x()) ** 2
                + (eventPos.y() - self._centre.y()) ** 2
            )
            self.prepareGeometryChange()
            self._radius = distance

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        if self._startStretch:
            self._startStretch = False
            self.setFlag(QGraphicsItem.ItemIsMovable, True)
            self.setFlag(QGraphicsItem.ItemIsSelectable, True)
            self._topLeft = self._centre - QPoint(self._radius, self._radius)
            self._rightBottom = self._centre + QPoint(self._radius, self._radius)
            self._end = self._centre + QPoint(self._radius, 0)
            self.setCursor(Qt.ArrowCursor)


class symbolArc(symbolShape):
    """
    Class to draw arc shapes. Can have four directions.
    """

    arcTypes = ["Up", "Right", "Down", "Left"]
    sides = ["Left", "Right", "Top", "Bottom"]

    def __init__(self, start: QPoint, end: QPoint):
        super().__init__()
        self._start = start
        self._end = end
        self._rect = QRectF(self._start, self._end).normalized()
        self._arcLine = QLineF(self._start, self._end)
        self._arcAngle = 0
        self._width = self._rect.width()
        self._height = self._rect.height()
        self._pen = symlyr.symbolPen
        self._adjustment = int(self._pen.width() / 2)
        self._stretchSide = None
        self._findAngle()

    def _findAngle(self):
        self._arcAngle = self._arcLine.angle()
        if 90 >= self._arcAngle >= 0:
            self._arcType = symbolArc.arcTypes[0]
        elif 180 >= self._arcAngle > 90:
            self._arcType = symbolArc.arcTypes[1]
        elif 270 >= self._arcAngle > 180:
            self._arcType = symbolArc.arcTypes[2]
        elif 360 > self._arcAngle > 270:
            self._arcType = symbolArc.arcTypes[3]

    def paint(self, painter, option, widget) -> None:
        if self.isSelected():
            painter.setPen(symlyr.selectedSymbolPen)
            painter.drawRect(self.rect)
            self.setZValue(symlyr.selectedSymbolLayer.z)
            if self._stretch:
                painter.setPen(symlyr.stretchSymbolPen)
                self.setZValue(symlyr.stretchSymbolLayer.z)
                if self._stretchSide == symbolArc.sides[0]:
                    painter.drawLine(self._rect.topLeft(), self._rect.bottomLeft())
                elif self._stretchSide == symbolArc.sides[1]:
                    painter.drawLine(self._rect.topRight(), self._rect.bottomRight())
                elif self._stretchSide == symbolArc.sides[2]:
                    painter.drawLine(self._rect.topLeft(), self._rect.topRight())
                elif self._stretchSide == symbolArc.sides[3]:
                    painter.drawLine(self._rect.bottomLeft(), self._rect.bottomRight())
        else:
            painter.setPen(symlyr.symbolPen)
            self.setZValue(symlyr.symbolLayer.z)

        self.arcDraw(painter)

    def arcDraw(self, painter):
        if self._arcType == symbolArc.arcTypes[0]:
            painter.drawArc(self._rect, 0, 180 * 16)
        elif self._arcType == symbolArc.arcTypes[1]:
            painter.drawArc(self._rect, 90 * 16, 180 * 16)
        elif self._arcType == symbolArc.arcTypes[2]:
            painter.drawArc(self._rect, 180 * 16, 180 * 16)
        elif self._arcType == symbolArc.arcTypes[3]:
            painter.drawArc(self._rect, 270 * 16, 180 * 16)

    def boundingRect(self):
        return self._rect.adjusted(-2, -2, 2, 2)

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, arc_rect: QRect):
        assert isinstance(arc_rect, QRect)
        self._rect = arc_rect.normalized()

    def __repr__(self):
        return f"symbolArc({self._start},{self._end})"

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, point: QPoint):
        assert isinstance(point, QPoint)
        self.prepareGeometryChange()
        self._start = (point,)

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, point: QPoint):
        assert isinstance(point, QPoint)
        self.prepareGeometryChange()
        self._end = point
        self._arcLine = QLineF(self._start, self._end)
        self._arcAngle = self._arcLine.angle()
        self._findAngle()
        self._rect = QRectF(self._start, self._end).normalized()

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width):
        self._width = width
        self.prepareGeometryChange()
        self.rect.setWidth(self._width)

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        self._height = height
        self.prepareGeometryChange()
        self.rect.setHeight(self._height)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        eventPos = event.pos().toPoint()
        if self._stretch:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            if eventPos.x() == self._rect.left():
                if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                    self.setCursor(Qt.SizeHorCursor)
                    self._stretchSide = symbolArc.sides[0]
            elif eventPos.x() == self._rect.right():
                if self._rect.top() <= eventPos.y() <= self._rect.bottom():
                    self.setCursor(Qt.SizeHorCursor)
                    self._stretchSide = symbolArc.sides[1]
            elif eventPos.y() == self._rect.top():
                if self._rect.left() <= eventPos.x() <= self._rect.right():
                    self.setCursor(Qt.SizeVerCursor)
                    self._stretchSide = symbolArc.sides[2]
            elif eventPos.y() == self._rect.bottom():
                if self._rect.left() <= eventPos.x() <= self._rect.right():
                    self.setCursor(Qt.SizeVerCursor)
                    self._stretchSide = symbolArc.sides[3]

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseMoveEvent(event)

        eventPos = event.pos().toPoint()
        if self._stretch:
            self.prepareGeometryChange()
            if self._stretchSide == symbolArc.sides[0]:
                self.setCursor(Qt.SizeHorCursor)
                self._rect.setLeft(eventPos.x())
            elif self._stretchSide == symbolArc.sides[1]:
                self.setCursor(Qt.SizeHorCursor)
                self._rect.setRight(eventPos.x() - self._adjustment)
            elif self._stretchSide == symbolArc.sides[2]:
                self.setCursor(Qt.SizeVerCursor)
                self._rect.setTop(eventPos.y())
            elif self._stretchSide == symbolArc.sides[3]:
                self.setCursor(Qt.SizeVerCursor)
                self._rect.setBottom(eventPos.y() - self._adjustment)
            self.update()

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        super().mouseReleaseEvent(event)
        if self.stretch:
            self._stretch = False
            self._stretchSide = None
            self.setCursor(Qt.ArrowCursor)

            if self._arcType == symbolArc.arcTypes[0]:
                self._start = self._rect.bottomLeft()
                self._end = self._rect.topRight()
            elif self._arcType == symbolArc.arcTypes[1]:
                self._start = self._rect.topLeft()
                self._end = self._rect.bottomRight()
            elif self._arcType == symbolArc.arcTypes[2]:
                self._start = self._rect.topRight()
                self._end = self._rect.bottomLeft()
            elif self._arcType == symbolArc.arcTypes[3]:
                self._start = self._rect.bottomRight()
                self._end = self._rect.topLeft()
            self._rect = QRectF(self._start, self._end).normalized()


class symbolLine(symbolShape):
    """
    line class definition for symbol drawing.
    """

    stretchSides = ["start", "end"]

    def __init__(
        self,
        start: QPoint,
        end: QPoint,
    ):
        super().__init__()
        self._end = end
        self._start = start
        self._stretch = False
        self._stretchSide = None
        self._pen = symlyr.symbolPen
        self._line = QLine(self._start, self._end)
        self._rect = QRect(self._start, self._end).normalized()
        self._horizontal = True  # True if line is horizontal, False if vertical

    def boundingRect(self):
        return self._rect.adjusted(-2, -2, 2, 2)

    def shape(self):
        path = QPainterPath()
        path.addRect(self._rect.adjusted(-2, -2, 2, 2))
        return path

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(symlyr.selectedSymbolPen)
            self.setZValue(symlyr.symbolLayer.z)
            if self._stretch:
                painter.setPen(symlyr.stretchSymbolPen)
                self.setZValue(symlyr.stretchSymbolLayer.z)
                if self._stretchSide == symbolLine.stretchSides[0]:
                    painter.drawEllipse(self._start, 2, 2)
                elif self._stretchSide == symbolLine.stretchSides[1]:
                    painter.drawEllipse(self._end, 2, 2)
        else:
            painter.setPen(symlyr.symbolPen)
            self.setZValue(symlyr.symbolLayer.z)
        painter.drawLine(self._line)

    def __repr__(self):
        return f"symbolLine({self._start},{self._end})"

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, rect: QRect):
        self._rect = rect

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start: QPoint):
        self.prepareGeometryChange()
        self._start = start
        self._line = QLine(self._start, self._end)
        self._rect = QRect(self._start, self._end).normalized()

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end: QPoint):
        self.prepareGeometryChange()
        self._end = end
        self._line = QLine(self._start, self._end)
        self._rect = QRect(self._start, self._end).normalized()

    @property
    def width(self):
        return self._pen.width()

    @width.setter
    def width(self, width: int):
        self._pen.setWidth(width)

    def bBox(self) -> QRect:
        return self.boundingRect()

    def Move(self, offset: QPoint):
        self.start += offset
        self._end += offset

    @property
    def length(self):
        return math.sqrt(
            (self.start.x() - self._end.x()) ** 2
            + (self.start.y() - self._end.y()) ** 2
        )

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        if self.isSelected() and self._stretch:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            eventPos = event.pos().toPoint()
            if eventPos == self.start:
                self._stretchSide = symbolLine.stretchSides[0]
            elif eventPos == self._end:
                self._stretchSide = symbolLine.stretchSides[1]

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        eventPos = event.pos().toPoint()
        if self._stretchSide == symbolLine.stretchSides[0]:
            self.prepareGeometryChange()
            self.start = eventPos
            self._line = QLine(self.start, self._end)
            self._rect = QRect(self.start, self._end).normalized()
        elif self._stretchSide == symbolLine.stretchSides[1]:
            self.prepareGeometryChange()
            self._end = eventPos
            self._line = QLine(self.start, self._end)
            self._rect = QRect(self.start, self._end).normalized()

        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        self._stretch = False
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self._stretchSide = ""


class symbolPolygon(symbolShape):
    def __init__(self, points: list):
        super().__init__()
        self._points = points
        self._polygon = QPolygonF(self._points)
        self._selectedCorner = None
        self._selectedCornerIndex = None

    def __repr__(self):
        return f"symbolPolygon({self._points})"

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(symlyr.selectedSymbolPen)
            self.setZValue(symlyr.selectedSymbolLayer.z)
            if self._selectedCorner is not None:
                painter.drawEllipse(self._selectedCorner, 5, 5)
        else:
            painter.setPen(symlyr.symbolPen)
            self.setZValue(symlyr.symbolLayer.z)
        painter.drawPolygon(self._polygon)

    def boundingRect(self) -> QRectF:
        return self._polygon.boundingRect()

    @property
    def polygon(self):
        return self._polygon

    @property
    def points(self) -> list:
        return self._points

    @points.setter
    def points(self, value: list):
        self.prepareGeometryChange()
        self._points = value
        self._polygon = QPolygonF(self._points)

    def addPoint(self, point: QPoint):
        self.prepareGeometryChange()
        self._points.append(point)
        self._polygon = QPolygonF(self._points)

    @property
    def tempLastPoint(self):
        return self._points[-1]

    @tempLastPoint.setter
    def tempLastPoint(self, value: QPoint):
        self.prepareGeometryChange()
        self._polygon = QPolygonF([*self._points, value])

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        eventPos = event.pos().toPoint()
        if self._stretch:
            self.setFlag(QGraphicsItem.ItemIsMovable, False)
            for point in self._points:
                if (eventPos - point).manhattanLength() <= self.scene().snapDistance:
                    self._selectedCorner = point
                    self._selectedCornerIndex = self._points.index(point)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        eventPos = event.pos().toPoint()
        if self._stretch:
            if self._selectedCorner is not None:
                self._points[self._selectedCornerIndex] = eventPos
                self.points = self._points
        else:
            super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        if self.stretch:
            self._stretch = False
            self._stretchSide = None
            self.setCursor(Qt.ArrowCursor)


class symbolPin(symbolShape):
    """
    symbol pin class definition for symbol drawing.
    """

    pinDirs = ["Input", "Output", "Inout"]
    pinTypes = ["Signal", "Ground", "Power", "Clock", "Digital", "Analog"]

    def __init__(
        self,
        start: QPoint,
        pinName: str,
        pinDir: str,
        pinType: str,
    ):
        super().__init__()

        self._start = start  # centre of pin
        self._pinName = pinName
        self._pinDir = pinDir
        self._pinType = pinType
        self._connected = False  # True if the pin is connected to a net.
        self._rect = QRect(self._start.x() - 5, self._start.y() - 5, 10, 10)

    def __str__(self):
        return f"symbolPin: {self._pinName} {self.mapToScene(self._start)}"

    def boundingRect(self):
        return self._rect  #

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(symlyr.selectedSymbolPinPen)
            painter.setBrush(symlyr.selectedSymbolPinBrush)
        else:
            painter.setPen(symlyr.symbolPinPen)
            painter.setBrush(symlyr.symbolPinBrush)
        painter.setFont(QFont("Arial", 12))
        painter.drawRect(self._rect)
        painter.drawText(
            QPoint(self._start.x() - 5, self._start.y() - 10), self._pinName
        )

    def __repr__(self):
        return f"pin({self._start},{self._pinName}, {self._pinDir}, {self._pinType})"

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        self.prepareGeometryChange()
        self._start = start
        self._rect = QRect(self._start.x() - 5, self._start.y() - 5, 10, 10)

    @property
    def pinName(self):
        return self._pinName

    @pinName.setter
    def pinName(self, pinName):
        if pinName != "":
            self.prepareGeometryChange()
            self._pinName = pinName

    @property
    def pinDir(self):
        return self._pinDir

    @pinDir.setter
    def pinDir(self, direction: str):
        if direction in symbolPin.pinDirs:
            self._pinDir = direction

    @property
    def pinType(self):
        return self._pinType

    @pinType.setter
    def pinType(self, pintype: str):
        if pintype in self.pinTypes:
            self._pinType = pintype

    @property
    def connected(self):
        return self._connected

    @connected.setter
    def connected(self, value: bool):
        if isinstance(value, bool):
            self._connected = value

    def toSchematicPin(self, start: QPoint):
        return schematicPin(start, self.pinName, self.pinDir, self.pinType)


class text(symbolShape):
    """
    This class is for text annotations on symbol or schematics.
    """

    textAlignments = ["Left", "Center", "Right"]
    textOrients = ["R0", "R90", "R180", "R270"]

    def __init__(
        self,
        start: QPoint,
        textContent: str,
        fontFamily: str,
        fontStyle: str,
        textHeight: str,
        textAlign: str,
        textOrient: str,
    ):
        super().__init__()
        self._start = start
        self._textContent = textContent
        self._textHeight = textHeight
        self._textAlign = textAlign
        self._textOrient = textOrient
        self._textFont = QFont(fontFamily)
        self._textFont.setStyleName(fontStyle)
        self._textFont.setPointSize(int(float(self._textHeight)))
        self._textFont.setKerning(True)
        self.setOpacity(1)
        self._fm = QFontMetrics(self._textFont)
        self._textOptions = QTextOption()
        self.setOrient()
        if self._textAlign == text.textAlignments[0]:
            self._textOptions.setAlignment(Qt.AlignmentFlag.AlignLeft)
        elif self._textAlign == text.textAlignments[1]:
            self._textOptions.setAlignment(Qt.AlignmentFlag.AlignCenter)
        elif self._textAlign == text.textAlignments[2]:
            self._textOptions.setAlignment(Qt.AlignmentFlag.AlignRight)
        self._rect = self._fm.boundingRect(
            QRect(0, 0, 400, 400), Qt.AlignmentFlag.AlignCenter, self._textContent
        )

    def __repr__(self):
        return (
            f"text({self._start},{self._textContent}, {self._textFont.family()},"
            f" {self._textFont.style()}, {self._textHeight}, {self._textAlign},"
            f"{self._textOrient})"
        )

    def setOrient(self):
        if self._textOrient == text.textOrients[0]:
            self.setRotation(0)
        elif self._textOrient == text.textOrients[1]:
            self.setRotation(90)
        elif self._textOrient == text.textOrients[2]:
            self.setRotation(180)
        elif self._textOrient == text.textOrients[3]:
            self.setRotation(270)
        elif self._textOrient == text.textOrients[4]:
            self.flip("x")
        elif self._labelOrient == text.textOrients[5]:
            self.flip("x")
            self.setRotation(90)
        elif self._labelOrient == text.textOrients[6]:
            self.flip("y")
            self.setRotation(90)

    def flip(self, direction: str):
        currentTransform = self.transform()
        newTransform = QTransform()
        if direction == "x":
            currentTransform = newTransform.scale(-1, 1) * currentTransform
        elif direction == "y":
            currentTransform = newTransform.scale(1, -1) * currentTransform
        self.setTransform(currentTransform)

    def boundingRect(self):
        if self._textAlign == text.textAlignments[0]:
            self._rect = self._fm.boundingRect(
                QRect(0, 0, 400, 400), Qt.AlignmentFlag.AlignLeft, self._textContent
            )
        elif self._textAlign == text.textAlignments[1]:
            self._rect = self._fm.boundingRect(
                QRect(0, 0, 400, 400), Qt.AlignmentFlag.AlignCenter, self._textContent
            )
        elif self._textAlign == text.textAlignments[2]:
            self._rect = self._fm.boundingRect(
                QRect(0, 0, 400, 400), Qt.AlignmentFlag.AlignRight, self._textContent
            )
        return QRect(
            self._start.x(),
            self._start.y() - self._rect.height(),
            self._rect.width(),
            self._rect.height(),
        )

    def paint(self, painter, option, widget):
        painter.setFont(self._textFont)
        if self.isSelected():
            painter.setPen(schlyr.selectedTextPen)
            painter.drawRect(self.boundingRect())
            self.setZValue(schlyr.selectedTextLayer.z)
        else:
            painter.setPen(schlyr.textPen)
            self.setZValue(schlyr.textLayer.z)
        painter.drawText(self.boundingRect(), self._textContent, o=self._textOptions)

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value: QPoint):
        self._start = value

    @property
    def textContent(self):
        return self._textContent

    @textContent.setter
    def textContent(self, inputText: str):
        if isinstance(inputText, str):
            self._textContent = inputText
        else:
            self.scene().logger.error(f"Not a string: {inputText}")

    @property
    def fontFamily(self) -> str:
        return self._textFont.family()

    @fontFamily.setter
    def fontFamily(self, familyName):
        fontFamilies = QFontDatabase.families(QFontDatabase.Latin)
        fixedFamilies = [
            family for family in fontFamilies if QFontDatabase.isFixedPitch(family)
        ]
        if familyName in fixedFamilies:
            self._textFont.setFamily(familyName)
        else:
            self.scene().logger.error(f"Not a valid font name: {familyName}")

    @property
    def fontStyle(self) -> str:
        return self._textFont.styleName()

    @fontStyle.setter
    def fontStyle(self, value: str):
        if value in QFontDatabase.styles(self._textFont.family()):
            self._textFont.setStyleName(value)
        else:
            self.scene().logger.error(f"Not a valid font style: {value}")

    @property
    def textHeight(self) -> str:
        return self._textHeight

    @textHeight.setter
    def textHeight(self, value: int):
        fontSizes = [
            str(size)
            for size in QFontDatabase.pointSizes(
                self._textFont.family(), self._textFont.styleName()
            )
        ]
        if value in fontSizes:
            self._textHeight = value
        else:
            self.scene().logger.error(f"Not a valid font height: {value}")
            self.scene().logger.warning(f"Valid font heights are: {fontSizes}")

    @property
    def textFont(self) -> QFont:
        return self._textFont

    @textFont.setter
    def textFont(self, value: QFont):
        assert isinstance(value, QFont)
        self._textFont = value

    @property
    def textAlignment(self):
        return self._textAlign

    @textAlignment.setter
    def textAlignment(self, value):
        if value in text.textAlignments:
            self._textAlign = value
        else:
            self.scene().logger.error(f"Not a valid text alignment value: {value}")

    @property
    def textOrient(self):
        return self._textOrient

    @textOrient.setter
    def textOrient(self, value):
        if value in text.textOrients:
            self._textOrient = value
        else:
            self.scene().logger.error(f"Not a valid text orientation: {value}")


class pinNetIndexTuple(NamedTuple):
    pin: symbolPin
    net: net.schematicNet
    netEndIndex: int


class schematicSymbol(symbolShape):
    def __init__(self, shapes: list, attr: dict):
        super().__init__()
        assert shapes is not None  # must not be an empty list
        self._shapes = shapes  # list of shapes in the symbol
        self._symattrs = attr  # parameters common to all instances of symbol
        self._counter = 0  # item's number on schematic
        self._libraryName = ""
        self._cellName = ""
        self._viewName = ""
        self._instanceName = ""
        self._netlistLine = ""
        self._labels: dict[str, symbolLabel] = dict()  # dict of labels
        self._pins: dict[str, symbolPin] = dict()  # dict of pins
        self._netlistIgnore: bool = False
        self._draft: bool = False
        self._pinLocations: dict[str, Union[QRect, QRectF]] = dict()  # pinName: pinRect
        self.pinNetMap: dict[str, str] = dict()  # pinName: netName
        self._pinNetIndexTupleSet: set[pinNetIndexTuple] = set()
        self._snapLines: dict[symbolPin, set[net.schematicNet]] = dict()
        self.addShapes()
        self.setFiltersChildEvents(True)
        self.setHandlesChildEvents(True)
        self.setFlag(QGraphicsItem.ItemContainsChildrenInShape, True)

        self.dashLines = dict()

    def addShapes(self):
        for item in self._shapes:
            item.setFlag(QGraphicsItem.ItemIsSelectable, False)
            item.setFlag(QGraphicsItem.ItemStacksBehindParent, True)
            item.setParentItem(self)
            if type(item) is symbolPin:
                self._pins[item.pinName] = item
            elif type(item) is symbolLabel:
                self._labels[item.labelName] = item

    def __repr__(self):
        return f"schematicSymbol({self._shapes})"

    def paint(self, painter, option, widget):
        self.setZValue(symlyr.symbolLayer.z)
        if self._draft:
            painter.setPen(symlyr.draftPen)
            self.setZValue(symlyr.draftLayer.z)
        if self.isSelected():
            painter.setPen(symlyr.selectedSymbolPen)
            painter.drawRect(self.borderRect)
            self.setZValue(symlyr.selectedSymbolLayer.z)
        if self.netlistIgnore:
            painter.setPen(schlyr.ignoreSymbolPen)
            painter.drawLine(self.borderRect.bottomLeft(), self.borderRect.topRight())
            painter.drawLine(self.borderRect.topLeft(), self.borderRect.bottomRight())

    def boundingRect(self):
        return self.childrenBoundingRect()

    def sceneEvent(self, event):
        try:  # if net is being drawn, do not accept any event.
            if self.scene().editModes.drawWire:
                return False
            else:
                super().sceneEvent(event)
                return True
        except AttributeError:
            return False

    def findPinNetIndexTuples(self):
        # Create an empty set to store pin-net-index tuples
        self._pinNetIndexTupleSet = set()

        # Iterate over each pin in the collection
        for pinItem in self._pins.values():
            # Find all the net items connected to the pin
            netsConnected = [
                netItem
                for netItem in self.scene().items(pinItem.sceneBoundingRect())
                if isinstance(netItem, net.schematicNet)
                and pinItem.mapToScene(pinItem.start).toPoint()
                in netItem.sceneEndPoints
            ]

            # Iterate over each connected net item
            for netItem in netsConnected:
                # Find the index of the pin in the net item's scene end points
                endIndex = netItem.sceneEndPoints.index(
                    pinItem.mapToScene(pinItem.start).toPoint()
                )

                # Create a pin-net-index tuple and add it to the set
                self._pinNetIndexTupleSet.add(
                    pinNetIndexTuple(pinItem, netItem, endIndex)
                )

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        self._snapLines = dict()
        self.findPinNetIndexTuples()
        for item in self._pinNetIndexTupleSet:
            self._snapLines.setdefault(item.pin, set())
            snapLine = net.guideLine(
                item.pin.mapToScene(item.pin.start),
                item.net.sceneEndPoints[item.netEndIndex - 1],
            )
            if item.net.nameSet:
                snapLine.name = item.net.name
                snapLine.nameSet = True
            self.scene().addItem(snapLine)
            self._snapLines[item.pin].add(snapLine)
            self.scene().removeItem(item.net)
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        for pin, snapLinesSet in self._snapLines.items():
            for snapLine in snapLinesSet:
                snapLine.setLine(
                    QLineF(
                        snapLine.mapFromScene(pin.mapToScene(pin.start)),
                        snapLine.line().p2(),
                    )
                )
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)

    @property
    def libraryName(self):
        return self._libraryName

    @libraryName.setter
    def libraryName(self, value):
        self._libraryName = value

    @property
    def cellName(self):
        return self._cellName

    @cellName.setter
    def cellName(self, value: str):
        self._cellName = value

    @property
    def viewName(self):
        return self._viewName

    @viewName.setter
    def viewName(self, value: str):
        self._viewName = value

    @property
    def instanceName(self):
        return self._instanceName

    @instanceName.setter
    def instanceName(self, value: str):
        """
        If instance name is changed and [@instName] label exists, change it too.
        """
        self._instanceName = value
        if self.labels.get("instanceName", None):
            self.labels["instanceName"].labelValue = value
            self.labels["instanceName"].labelValueSet = True
            self.labels["instanceName"].update()

    @property
    def counter(self) -> int:
        return self._counter

    @counter.setter
    def counter(self, value: int):
        assert isinstance(value, int)
        self._counter = value

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value: float):
        self.setRotation(value)
        self._angle = value

    @property
    def labels(self):
        return self._labels  # dictionary

    @cached_property
    def pins(self):
        if self.symattrs.get("pinOrder"):
            pinOrderList = self.symattrs.get("pinOrder").split(",")
            orderedPins = {key.strip(): self._pins[key.strip()] for key in pinOrderList}
            return orderedPins
        else:
            return self._pins

    @property
    def shapes(self):
        return self._shapes

    @shapes.setter
    def shapes(self, shapeList: list):
        self.prepareGeometryChange()
        self._shapes = shapeList
        self.addShapes()

    @property
    def symattrs(self):
        return self._symattrs

    @symattrs.setter
    def symattrs(self, attrDict: dict):
        self._symattrs = attrDict

    @property
    def netlistIgnore(self) -> bool:
        return self._netlistIgnore

    @netlistIgnore.setter
    def netlistIgnore(self, value: bool):
        assert isinstance(value, bool)
        self._netlistIgnore = value

    @property
    def borderRect(self):
        return self.childrenBoundingRect().normalized().adjusted(-1, -1, 1, 1)


class schematicPin(symbolShape):
    """
    schematic pin class.
    """

    pinDirs = ["Input", "Output", "Inout"]
    pinTypes = ["Signal", "Ground", "Power", "Clock", "Digital", "Analog"]

    def __init__(
        self,
        start: QPoint,
        pinName: str,
        pinDir: str,
        pinType: str,
    ):
        super().__init__()
        self._start = start
        self._pinName = pinName
        self._pinDir = pinDir
        self._pinType = pinType
        self._pinNetIndexTupleSet = set()

    def __repr__(self):
        return (
            f"schematicPin({self._start}, {self._pinName}, {self._pinDir}, "
            f"{self._pinType})"
        )

    def paint(self, painter, option, widget):
        if self.isSelected():
            painter.setPen(schlyr.selectedSchematicPinPen)
            painter.setBrush(schlyr.selectedSchematicPinBrush)
            painter.drawRect(
                QRect.span(
                    QPoint(self._start.x() - 10, self._start.y() - 10),
                    QPoint(self._start.x() + 10, self._start.y() + 10),
                )
            )
        else:
            painter.setPen(schlyr.schematicPinPen)
            painter.setBrush(schlyr.schematicPinBrush)
            self.setZValue(schlyr.schematicPinLayer.z)
        self.drawPin(painter)

    def drawPin(self, painter):
        painter.setFont(QFont("Arial", 12))
        match self.pinDir:
            case "Input":
                painter.drawPolygon(
                    [
                        QPoint(self._start.x() - 10, self._start.y() - 10),
                        QPoint(self._start.x() + 10, self._start.y() - 10),
                        QPoint(self._start.x() + 20, self._start.y()),
                        QPoint(self._start.x() + 10, self._start.y() + 10),
                        QPoint(self._start.x() - 10, self._start.y() + 10),
                    ]
                )
            case "Output":
                painter.drawPolygon(
                    [
                        QPoint(self._start.x() - 20, self._start.y()),
                        QPoint(self._start.x() - 10, self._start.y() - 10),
                        QPoint(self._start.x() + 10, self._start.y() - 10),
                        QPoint(self._start.x() + 10, self._start.y() + 10),
                        QPoint(self._start.x() - 10, self._start.y() + 10),
                    ]
                )
            case "Inout":
                painter.drawPolygon(
                    [
                        QPoint(self._start.x() - 20, self._start.y()),
                        QPoint(self._start.x() - 10, self._start.y() - 10),
                        QPoint(self._start.x() + 10, self._start.y() - 10),
                        QPoint(self._start.x() + 20, self._start.y()),
                        QPoint(self._start.x() + 10, self._start.y() + 10),
                        QPoint(self._start.x() - 10, self._start.y() + 10),
                    ]
                )
        painter.drawText(self._start.x(), self._start.y() - 20, self.pinName)

    def boundingRect(self):
        return QRect(self._start.x() - 10, self._start.y() - 10, 30, 20).adjusted(
            -5, -10, 5, 5
        )

    def sceneEvent(self, event):
        if self.scene().editModes.drawWire:
            return False
        else:
            super().sceneEvent(event)
            return True

    def findPinNetIndexTuples(self):
        # Create an empty set to store pin-net-index tuples
        self._pinNetIndexTupleSet = set()

        pinSceneConnectRect = self.mapRectToScene(
            QRect.span(
                QPoint(self._start.x() - 5, self._start.y() - 5),
                QPoint(self._start.x() + 5, self._start.y() + 5),
            )
        ).toRect()
        # Find all the net items connected to the pin
        netsConnected = [
            netItem
            for netItem in self.scene().items(pinSceneConnectRect)
            if isinstance(netItem, net.schematicNet)
        ]

        # Iterate over each connected net item
        for netItem in netsConnected:
            netEndIndex = list(
                map(pinSceneConnectRect.contains, netItem.sceneEndPoints)
            ).index(True)
            self._pinNetIndexTupleSet.add(pinNetIndexTuple(self, netItem, netEndIndex))

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)
        self.findPinNetIndexTuples()
        for tupleItem in self._pinNetIndexTupleSet:
            self._snapLines = set()
            snapLine = net.guideLine(
                self.mapToScene(self.start),
                tupleItem.net.sceneEndPoints[tupleItem.netEndIndex - 1],
            )
            if tupleItem.net.nameSet:
                snapLine.name = tupleItem.net.name
                snapLine.nameSet = True
            self.scene().addItem(snapLine)
            self._snapLines.add(snapLine)
            self.scene().removeItem(tupleItem.net)
        super().mousePressEvent(event)

    #
    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        lines = []
        for snapLine in self._snapLines:
            lines = self.scene().addStretchWires(
                self.mapToScene(self.start).toPoint(),
                snapLine.mapToScene(snapLine.line().p2()).toPoint(),
            )
            self.scene().removeItem(snapLine)
        if lines:
            for line in lines:
                if snapLine.nameSet:
                    line.name = snapLine.name
                    line.nameSet = True
            self.scene().addListUndoStack(lines)
        self._snapLines = dict()

    def itemChange(self, change, value):
        # if self.scene() and change == QGraphicsItem.ItemPositionHasChanged:
        #     # item's position has changed
        #     # do something here
        #     for item in self._netTupleSet:
        #         if item.start == 1:
        #             item.net.start = self.mapToScene(self.start)
        #         elif item.start == 0:
        #             item.net.end = self.mapToScene(self.start)
        return super().itemChange(change, value)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseMoveEvent(event)
        for snapLine in self._snapLines:
            snapLine.setLine(
                QLineF(
                    snapLine.mapFromScene(self.mapToScene(self.start)),
                    snapLine.line().p2(),
                )
            )

    def toSymbolPin(self, start: QPoint):
        return symbolPin(start, self.pinName, self.pinDir, self.pinType)

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        self._start = start

    @property
    def pinName(self):
        return self._pinName

    @pinName.setter
    def pinName(self, pinName):
        if pinName != "":
            self._pinName = pinName

    @property
    def pinDir(self):
        return self._pinDir

    @pinDir.setter
    def pinDir(self, direction: str):
        if direction in self.pinDirs:
            self._pinDir = direction

    @property
    def pinType(self):
        return self._pinType

    @pinType.setter
    def pinType(self, pintype: str):
        if pintype in self.pinTypes:
            self._pinType = pintype



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

# Load symbol and maybe later schematic from json file.
# import pathlib

import json
import pdk.process as fabproc
from PySide6.QtCore import QPoint, QLineF
from PySide6.QtWidgets import QGraphicsScene

from revedaEditor.common.net import schematicNet
import revedaEditor.common.shapes as shp
import revedaEditor.common.labels as lbl
import revedaEditor.common.layoutShapes as lshp
import revedaEditor.fileio.symbolEncoder as se
import pdk.layoutLayers as laylyr
import pdk.pcells as pcells
import pathlib


class symbolItems:
    def __init__(self, scene: QGraphicsScene):
        """
        Initializes the class instance.

        Args:
            scene (QGraphicsScene): The QGraphicsScene object.

        """
        self.scene = scene
        self.snapTuple = scene.snapTuple

    def create(self, item: dict):
        """
        Create symbol items from json file.
        """
        match item["type"]:
            case "rect":
                return self.createRectItem(item)
            case "circle":
                return self.createCircleItem(item)
            case "arc":
                return self.createArcItem(item)
            case "line":
                return self.createLineItem(item)
            case "pin":
                return self.createPinItem(item)
            case "label":
                return self.createLabelItem(item)
            case "text":
                return self.createTextItem(item)
            case "polygon":
                return self.createPolygonItem(item)

    def createRectItem(self, item: dict):
        """
        Create symbol items from json file.
        """
        start = QPoint(item["rect"][0], item["rect"][1])
        end = QPoint(item["rect"][2], item["rect"][3])
        rect = shp.symbolRectangle(start, end)
        rect.setPos(
            QPoint(item["loc"][0], item["loc"][1]),
        )
        rect.angle = item["ang"]
        return rect

    def createCircleItem(self, item: dict):
        centre = QPoint(item["cen"][0], item["cen"][1])
        end = QPoint(item["end"][0], item["end"][1])
        circle = shp.symbolCircle(centre, end)  # note that we are using grid
        # values for
        # scene
        circle.setPos(
            QPoint(item["loc"][0], item["loc"][1]),
        )
        circle.angle = item["ang"]
        return circle

    def createArcItem(self, item: dict):
        start = QPoint(item["st"][0], item["st"][1])
        end = QPoint(item["end"][0], item["end"][1])

        arc = shp.symbolArc(start, end)  # note that we are using grid values
        # for scene
        arc.setPos(QPoint(item["loc"][0], item["loc"][1]))
        arc.angle = item["ang"]
        return arc

    def createLineItem(self, item: dict):
        start = QPoint(item["st"][0], item["st"][1])
        end = QPoint(item["end"][0], item["end"][1])

        line = shp.symbolLine(start, end)
        line.setPos(QPoint(item["loc"][0], item["loc"][1]))
        line.angle = item["ang"]
        return line

    def createPinItem(self, item: dict):
        start = QPoint(item["st"][0], item["st"][1])
        pin = shp.symbolPin(start, item["nam"], item["pd"], item["pt"])
        pin.setPos(QPoint(item["loc"][0], item["loc"][1]))
        pin.angle = item["ang"]
        return pin

    def createLabelItem(self, item: dict):
        start = QPoint(item["st"][0], item["st"][1])
        label = lbl.symbolLabel(
            start,
            item["def"],
            item["lt"],
            item["ht"],
            item["al"],
            item["or"],
            item["use"],
        )
        label.setPos(QPoint(item["loc"][0], item["loc"][1]))
        label.labelName = item["nam"]
        label.labelText = item["txt"]
        label.labelVisible = item["vis"]
        label.labelValue = item["val"]
        return label

    def createTextItem(self, item: dict):
        start = QPoint(item["st"][0], item["st"][1])
        text = shp.text(
            start,
            item["tc"],
            item["ff"],
            item["fs"],
            item["th"],
            item["ta"],
            item["to"],
        )
        text.setPos(QPoint(item["loc"][0], item["loc"][1]))
        return text

    def createPolygonItem(self, item: dict):
        pointsList = [QPoint(point[0], point[1]) for point in item["ps"]]
        return shp.symbolPolygon(pointsList)

    @staticmethod
    def createSymbolAttribute(item: dict):
        return se.symbolAttribute(item["nam"], item["def"])


class schematicItems:
    def __init__(self, scene: QGraphicsScene):
        self.scene = scene
        self.libraryDict = scene.libraryDict
        self.snapTuple = scene.snapTuple

    def create(self, item: dict):
        match item["type"]:
            case "sys":
                itemShapes = list()
                symbolAttributes = dict()
                symbolInstance = shp.schematicSymbol(itemShapes, symbolAttributes)
                symbolInstance.libraryName = item["lib"]
                symbolInstance.cellName = item["cell"]
                symbolInstance.viewName = item["view"]
                symbolInstance.counter = item["ic"]
                symbolInstance.instanceName = item["nam"]
                symbolInstance.angle = item.get("ang", 0)
                symbolInstance.netlistIgnore = bool(item.get("ign", 0))
                symbolInstance.labelDict = item["ld"]
                symbolInstance.setPos(*item["loc"])
                [labelItem.labelDefs() for labelItem in symbolInstance.labels.values()]
                libraryPath = self.libraryDict.get(item["lib"])

                if libraryPath is None:
                    self.createDraftSymbol(item, symbolInstance)
                    self.scene.logger.warning(f"{item['lib']} cannot be found.")
                    return symbolInstance
                else:
                    # find the symbol file
                    file = libraryPath.joinpath(item["cell"], f'{item["view"]}.json')
                    if not file.exists():
                        self.createDraftSymbol(item, symbolInstance)
                        self.scene.logger.warning(f"{item['lib']} cannot be found.")
                        return symbolInstance
                    else:
                        # load json file and create shapes
                        with file.open(mode="r", encoding="utf-8") as temp:
                            try:
                                jsonItems = json.load(temp)
                                assert jsonItems[0]["cellView"] == "symbol"
                                symbolSnapTuple = jsonItems[1]["snapGrid"]
                                # we snap to scene grid values. Need to test further.
                                symbolShape = symbolItems(self.scene)
                                symbolShape.snapTuple = symbolSnapTuple
                                for jsonItem in jsonItems[
                                    2:
                                ]:  # skip first two entries.
                                    if jsonItem["type"] == "attr":
                                        symbolAttributes[jsonItem["nam"]] = jsonItem[
                                            "def"
                                        ]
                                    else:
                                        itemShapes.append(symbolShape.create(jsonItem))
                                symbolInstance.shapes = itemShapes
                                for labelItem in symbolInstance.labels.values():
                                    if (
                                        labelItem.labelName
                                        in symbolInstance.labelDict.keys()
                                    ):
                                        labelItem.labelValue = symbolInstance.labelDict[
                                            labelItem.labelName
                                        ][0]
                                        labelItem.labelVisible = (
                                            symbolInstance.labelDict[
                                                labelItem.labelName
                                            ][1]
                                        )
                                symbolInstance.symattrs = symbolAttributes
                                [
                                    labelItem.labelDefs()
                                    for labelItem in symbolInstance.labels.values()
                                ]
                                return symbolInstance
                            except json.decoder.JSONDecodeError:
                                self.scene.logger.error("Error: Invalid Symbol file")

            case "scn":
                start = QPoint(item["st"][0], item["st"][1])
                end = QPoint(item["end"][0], item["end"][1])
                netItem = schematicNet(start, end)
                netItem.name = item["nam"]
                netItem.nameSet = item["ns"]
                return netItem
            case "scp":
                start = QPoint(item["st"][0], item["st"][1])
                pinName = item["pn"]
                pinDir = item["pd"]
                pinType = item["pt"]
                pinItem = shp.schematicPin(
                    start,
                    pinName,
                    pinDir,
                    pinType,
                )
                pinItem.angle = item["ang"]
                return pinItem
            case "txt":
                start = QPoint(item["st"][0], item["st"][1])
                text = shp.text(
                    start,
                    item["tc"],
                    item["ff"],
                    item["fs"],
                    item["th"],
                    item["ta"],
                    item["to"],
                )
                return text
            case "dot":
                start = QPoint(item["pt"][0], item["pt"][1])
                dot = crossingDot(start)
                return dot

    def createDraftSymbol(self, item: dict, symbolInstance: shp.schematicSymbol):
        rectItem = shp.symbolRectangle(
            QPoint(item["br"][0], item["br"][1]), QPoint(item["br"][2], item["br"][3])
        )
        fixedFont = self.scene.fixedFont
        textItem = shp.text(
            rectItem.start,
            f'{item["lib"]}' f'/{item["cell"]}/' f'{item["view"]}',
            fixedFont.family(),
            fixedFont.styleName(),
            fixedFont.pointSize(),
            shp.text.textAlignments[0],
            shp.text.textOrients[0],
        )
        symbolInstance.shapes = [rectItem, textItem]
        symbolInstance.draft = True


class layoutItems:
    def __init__(self, scene: QGraphicsScene):
        """
        Create layout items from json file.
        """
        self.scene = scene
        self.libraryDict = scene.libraryDict
        self.rulerFont = scene.rulerFont
        self.rulerTickLength = scene.rulerTickLength
        self.snapTuple = scene.snapTuple
        self.rulerWidth = scene.rulerWidth
        self.rulerTickGap = scene.rulerTickGap

    def create(self, item: dict):
        match item["type"]:
            case "Inst":
                return self.createLayoutInstance(item)
            case "Pcell":
                libraryPath = pathlib.Path(self.libraryDict.get(item["lib"]))
                if libraryPath is None:
                    self.scene.logger.error(f'{item["lib"]} cannot be found.')
                    return None
                cell = item["cell"]
                viewName = item["view"]
                # open pcell json file with reference to pcell class name
                file = libraryPath.joinpath(cell, f"{viewName}.json")
                with file.open("r") as temp:  # open pcell view item
                    try:
                        pcellDef = json.load(temp)
                        if pcellDef[0]["cellView"] != "pcell":
                            self.scene.logger.error("Not a pcell cell")
                        else:
                            pcellInstance = eval(f'pcells.{pcellDef[1]["reference"]}()')
                            pcellInstance(**item["params"])
                            pcellInstance.libraryName = item["lib"]
                            pcellInstance.cellName = item["cell"]
                            pcellInstance.viewName = item["view"]
                            pcellInstance.counter = item["ic"]
                            pcellInstance.instanceName = item["nam"]
                            pcellInstance.setPos(QPoint(item["loc"][0], item["loc"][1]))
                            return pcellInstance
                    except json.decoder.JSONDecodeError:
                        print("Error: Invalid PCell file")
            case "Rect":
                return self.createRectShape(item)
            case "Path":
                return self.createPathShape(item)
            case "Label":
                return self.createLabelShape(
                    item,
                )
            case "Pin":
                return self.createPinShape(item)
            case "Polygon":
                return self.createPolygonShape(item)
            case "Via":
                return self.createViaArrayShape(item)
            case "Ruler":
                return self.createRulerShape(item)

    def createLayoutInstance(self, item):
        libraryPath = pathlib.Path(self.libraryDict.get(item["lib"]))
        if libraryPath is None:
            print(f'{item["lib"]} cannot be found.')
            return None
        cell = item["cell"]
        viewName = item["view"]
        instCounter = item["ic"]
        file = libraryPath.joinpath(cell, f"{viewName}.json")
        itemShapes = list()
        with open(file, "r") as temp:
            try:
                shapes = json.load(temp)
                for shape in shapes[2:]:
                    itemShapes.append(layoutItems(self.scene).create(shape))
            except json.decoder.JSONDecodeError:
                print("Error: Invalid Layout file")
        layoutInstance = lshp.layoutInstance(itemShapes)
        layoutInstance.libraryName = item["lib"]
        layoutInstance.cellName = item["cell"]
        layoutInstance.counter = instCounter
        layoutInstance.instanceName = item.get("nam", "")
        layoutInstance.setPos(item["loc"][0], item["loc"][1])
        layoutInstance.angle = item.get("ang", 0)
        layoutInstance.viewName = viewName
        return layoutInstance

    def createRectShape(self, item):
        start = QPoint(item["tl"][0], item["tl"][1])
        end = QPoint(item["br"][0], item["br"][1])
        layoutLayer = laylyr.pdkDrawingLayers[item["ln"]]
        rect = lshp.layoutRect(start, end, layoutLayer)
        # rect.setPos(QPoint(item["loc"][0], item["loc"][1]))
        rect.angle = item.get("ang", 0)
        return rect

    def createPathShape(self, item):
        path = lshp.layoutPath(
            QLineF(
                QPoint(item["dfl1"][0], item["dfl1"][1]),
                QPoint(item["dfl2"][0], item["dfl2"][1]),
            ),
            laylyr.pdkDrawingLayers[item["ln"]],
            item["w"],
            item["se"],
            item["ee"],
            item["md"],
        )
        path.name = item.get("nam", "")
        path.angle = item.get("ang", 0)
        return path

    def createRulerShape(self, item):
        ruler = lshp.layoutRuler(
            QLineF(
                QPoint(item["dfl1"][0], item["dfl1"][1]),
                QPoint(item["dfl2"][0], item["dfl2"][1]),
            ),
            self.rulerWidth,
            self.rulerTickGap,
            self.rulerTickLength,
            self.rulerFont,
            item["md"],
        )
        ruler.angle = item.get("ang", 0)
        return ruler

    def createLabelShape(self, item):
        layoutLayer = laylyr.pdkTextLayers[item["ln"]]
        label = lshp.layoutLabel(
            QPoint(item["st"][0], item["st"][1]),
            item["lt"],
            item["ff"],
            item["fs"],
            item["fh"],
            item["la"],
            item["lo"],
            layoutLayer,
        )
        label.angle = item.get("ang", 0)
        return label

    def createPinShape(self, item):
        layoutLayer = laylyr.pdkPinLayers[item["ln"]]
        pin = lshp.layoutPin(
            QPoint(item["tl"][0], item["tl"][1]),
            QPoint(item["br"][0], item["br"][1]),
            item["pn"],
            item["pd"],
            item["pt"],
            layoutLayer,
        )
        pin.angle = item.get("ang", 0)
        return pin

    def createPolygonShape(self, item):
        layoutLayer = laylyr.pdkDrawingLayers[item["ln"]]
        pointsList = [QPoint(point[0], point[1]) for point in item["ps"]]
        polygon = lshp.layoutPolygon(pointsList, layoutLayer)
        polygon.angle = item.get("ang", 0)
        return polygon

    def createViaArrayShape(self, item):
        viaDefTuple = fabproc.processVias[
            fabproc.processViaNames.index(item["via"]["vdt"])
        ]
        via = lshp.layoutVia(
            QPoint(item["via"]["st"][0], item["via"]["st"][1]),
            viaDefTuple,
            item["via"]["w"],
            item["via"]["h"],
        )
        viaArray = lshp.layoutViaArray(
            QPoint(item["st"][0], item["st"][1]),
            via,
            item["sp"],
            item["xn"],
            item["yn"],
        )
        viaArray.angle = item.get("ang", 0)
        return viaArray



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#    Add-ons and extensions developed for this software may be distributed
#    under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

# Compilation mode, standalone everywhere, except on macOS there app bundle
# nuitka-project-if: {OS} in ("Windows", "Linux", "FreeBSD"):
#    nuitka-project: --onefile
# nuitka-project-if: {OS} == "Darwin":
#    nuitka-project: --standalone
#    nuitka-project: --macos-create-app-bundle
#
# Debugging options, controlled via environment variable at compile time.
# nuitka-project-if: os.getenv("DEBUG_COMPILATION", "no") == "yes":
#     nuitka-project: --enable-console
# nuitka-project-else:
#     nuitka-project: --disable-console
# pyside6 plugin
# nuitka-project: --enable-plugin=pyside6
# nuitka-project: --include-data-files=./.env=.env


import os
import sys
from PySide6.QtWidgets import QApplication

# from PySide6.QtGui import QScreen, QGuiApplication
import revedaEditor.gui.revedaMain as rvm
import revedaEditor.gui.pythonConsole as pcon
from contextlib import redirect_stdout, redirect_stderr
from dotenv import load_dotenv
from pathlib import Path


# simulation window


class revedaApp(QApplication):
    """
    Initializes the class instance and sets the paths to the revedaeditor and revedasim if the corresponding
    environment variables exist. Also adds the paths to the system path if available.

    Returns:
        None
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Load environment variables
        load_dotenv()
        reveda_runpathObj = Path(__file__).resolve().parent
        self.revedaeditor_path = None
        self.revedasim_path = None
        self.reveda_pdk_path = None
        self.setPaths(reveda_runpathObj)

    def setPaths(self, reveda_runpathObj):
        # Set the paths to the revedaeditor and revedasim
        self.revedaeditor_path = os.environ.get("REVEDAEDIT_PATH", None)
        if self.revedaeditor_path:
            if Path(self.revedaeditor_path).is_absolute():
                self.revedaeditor_pathObj = Path(self.revedaeditor_path)
            else:
                self.revedaeditor_pathObj = reveda_runpathObj.joinpath(
                    self.revedaeditor_path
                )
        self.revedasim_path = os.environ.get("REVEDASIM_PATH", None)
        if self.revedasim_path:
            if Path(self.revedasim_path).is_absolute():
                self.revedasim_pathObj = Path(self.revedasim_path)
            else:
                self.revedasim_pathObj = reveda_runpathObj.joinpath(self.revedasim_path)
            sys.path.append(str(self.revedasim_pathObj))
        self.reveda_pdk_path = os.environ.get("REVEDA_PDK_PATH", None)
        if self.reveda_pdk_path:
            if Path(self.reveda_pdk_path).is_absolute():
                self.reveda_pdk_pathObj = Path(self.reveda_pdk_path)
            else:
                self.reveda_pdk_pathObj = reveda_runpathObj.joinpath(self.reveda_pdk_path)
            sys.path.append(str(self.reveda_pdk_pathObj))


def main():
    # Start Main application window
    app = revedaApp(sys.argv)

    mainW = rvm.MainWindow()
    mainW.setWindowTitle("Revolution EDA")

    # empty argument as there is no parent window.
    redirect = pcon.Redirect(mainW.centralW.console.errorwrite)
    with redirect_stdout(mainW.centralW.console), redirect_stderr(redirect):
        mainW.show()
        sys.exit(app.exec())


if __name__ == "__main__":
    main()



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


from PySide6.QtGui import QFont, QFontMetrics, QFontDatabase


class font(QFont):
    def __init__(
        self, fontFamily: str, fontStyle: str, fontSize: int, kerning: bool, logger
    ):
        self._fontFamily = fontFamily
        self._fontStyle = fontStyle
        self._fontSize = fontSize
        self._kerning = kerning
        self._logger = logger
        self.setFamily(self._fontFamily)
        self.setStyle(self._fontStyle)
        self.setPointSize(self._fontSize)
        self.setKerning(self._kerning)
        self.setStyleHint(QFont.TypeWriter)

    @property
    def fontFamily(self):
        return self._fontFamily

    @fontFamily.setter
    def fontFamily(self, value: str):
        fontFamilies = QFontDatabase.families(QFontDatabase.Latin)
        fixedFamilies = [
            family for family in fontFamilies if QFontDatabase.isFixedPitch(family)
        ]
        if value not in fixedFamilies:
            self._logger.warning(
                "No such font family present in the system. Another "
                "font "
                "will be substituted."
            )
        self._fontFamily = value
        self.setFamily(self._fontFamily)

    @property
    def fontStyle(self):
        return self._fontStyle

    @fontStyle.setter
    def fontStyle(self, value: str):
        fstyles = QFontDatabase.styles(self._fontFamily)
        if value not in fstyles:
            self._logger.warning(
                f"No matching style is available. Available styles "
                f'are: {",".join(fstyles)}'
            )
        else:
            self.setStyle(value)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

# This module includes all the base definitions for schematic drawings.
from dataclasses import replace

from PySide6.QtCore import (Qt)
from PySide6.QtGui import (QColor, QPen, QBrush)

import revedaEditor.backend.dataDefinitions as ddef

# schematic layers

wireLayer = ddef.edLayer(name="wire", pcolor=QColor("cyan"), pwidth=1,
                         pstyle=Qt.SolidLine, z=0, bcolor=QColor("cyan"),
                         bstyle=Qt.SolidPattern, visible=True, selectable=True)
wireErrorLayer = replace(wireLayer, name="wireError", pcolor=QColor("red"),
                         z=1)
selectedWireLayer = replace(wireLayer, name="selectedWire",
                            pcolor=QColor("blue"), z=2)
wireHilightLayer = ddef.edLayer(name='wireHilightLayer',
                                pcolor=QColor('darkMagenta'), pwidth=5, z=6,
                                visible=True, selectable=False)
guideLineLayer = replace(wireLayer, name="guideLine", pcolor=QColor("gray"),
                         pstyle=Qt.DashLine, z=7)
textLayer = ddef.edLayer(name="text", pcolor=QColor("white"), pwidth=1, z=4,
                         visible=True, selectable=True)

schematicPinLayer = ddef.edLayer(name="schematicPin", pcolor=QColor("red"),
                                 pwidth=1, z=3, bcolor=QColor("red"),
                                 bstyle=Qt.SolidPattern, visible=True,
                                 selectable=True)
selectedSchematicPinLayer = replace(schematicPinLayer,
                                    name="selectedSchematicPin",
                                    pcolor=QColor("yellow"), z=4)
selectedTextLayer = replace(textLayer, name="selectedText", pcolor=QColor("yellow"), z=5)
ignoreSymbolLayer = ddef.edLayer(name='ignoreLayer', pcolor=QColor('red'), pwidth=5,
                                 z=6, visible=True, selectable=False)
otherLayer = ddef.edLayer(name='otherLayer', pcolor=QColor('gray'), bcolor=QColor('gray'),
                          pwidth=1, z=0, visible=True,
                          selectable=False)

draftLayer = ddef.edLayer(name='draftLayer', pcolor=QColor('gray'), pwidth=1, z=0,
                          bcolor=QColor('gray'),
                          bstyle=Qt.DiagCrossPattern,
                          visible=True, selectable=True)
# schematic pens
schematicPinPen = QPen(schematicPinLayer.pcolor, schematicPinLayer.pwidth,
                       schematicPinLayer.pstyle)
selectedSchematicPinPen = QPen(selectedSchematicPinLayer.pcolor,
                               selectedSchematicPinLayer.pwidth,
                               selectedSchematicPinLayer.pstyle)
textPen = QPen(textLayer.pcolor, textLayer.pwidth, textLayer.pstyle)
selectedTextPen = QPen(selectedTextLayer.pcolor, selectedTextLayer.pwidth,
                       selectedTextLayer.pstyle)
guideLinePen = QPen(guideLineLayer.pcolor, guideLineLayer.pwidth,
                    guideLineLayer.pstyle)
wirePen = QPen(wireLayer.pcolor, wireLayer.pwidth, wireLayer.pstyle)

selectedWirePen = QPen(selectedWireLayer.pcolor,
                       selectedWireLayer.pwidth,
                       selectedWireLayer.pstyle)
stretchWirePen = QPen(QColor('red'), wireLayer.pwidth, wireLayer.pstyle)
errorWirePen = QPen(wireErrorLayer.pcolor, wireErrorLayer.pwidth, wireErrorLayer.pstyle)
ignoreSymbolPen = QPen(ignoreSymbolLayer.pcolor,
                       ignoreSymbolLayer.pwidth,
                       ignoreSymbolLayer.pstyle)
hilightPen = QPen(wireHilightLayer.pcolor, wireHilightLayer.pwidth,
                  wireHilightLayer.pstyle)
otherPen = QPen(otherLayer.pcolor, otherLayer.pwidth, otherLayer.pstyle)
draftPen = QPen(draftLayer.pcolor, draftLayer.pwidth, draftLayer.pstyle)

# schematic brushes
schematicPinBrush = QBrush(schematicPinLayer.bcolor, schematicPinLayer.bstyle)
wireBrush = QBrush(wireLayer.bcolor, wireLayer.bstyle)
selectedWireBrush = QBrush(selectedWireLayer.bcolor, selectedWireLayer.bstyle)
selectedSchematicPinBrush = QBrush(selectedSchematicPinLayer.bcolor,
                                   selectedSchematicPinLayer.bstyle)
otherBrush = QBrush(otherLayer.bcolor, otherLayer.bstyle)
draftBrush = QBrush(draftLayer.bcolor, draftLayer.bstyle)

# crossing dot diameter
crossingDotDiameter = 2



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

import json

import revedaEditor.common.net as net
import revedaEditor.common.shapes as shp


class schematicEncoder(json.JSONEncoder):
    def default(self, item):
        if isinstance(item, shp.schematicSymbol):
            # if item was drawn as a draft, then just carry the labels
            if item.draft:
                itemLabelDict = item.labelDict
            else:
                itemLabelDict = {
                    label.labelName: [label.labelValue, label.labelVisible]
                    for label in item.labels.values()
                }
            itemDict = {
                "type": "sys",
                "lib": item.libraryName,
                "cell": item.cellName,
                "view": item.viewName,
                "nam": item.instanceName,
                "ic": item.counter,
                "ld": itemLabelDict,
                "loc": (item.scenePos() - item.scene().origin).toTuple(),
                "ang": item.angle,
                "ign": int(item.netlistIgnore),
                "br": item.boundingRect().getCoords(),
            }
            return itemDict
        elif isinstance(item, net.schematicNet):
            itemDict = {
                "type": "scn",
                "st": (item.mapToScene(item.draftLine.p1()) - item.scene().origin)
                .toPoint()
                .toTuple(),
                "end": (item.mapToScene(item.draftLine.p2()) - item.scene().origin)
                .toPoint()
                .toTuple(),
                "nam": item.name,
                "ns": item.nameSet,
            }
            return itemDict
        elif isinstance(item, shp.schematicPin):
            itemDict = {
                "type": "scp",
                "st": (item.mapToScene(item.start) - item.scene().origin).toTuple(),
                "pn": item.pinName,
                "pd": item.pinDir,
                "pt": item.pinType,
                "ang": item.angle,
            }
            return itemDict
        elif isinstance(item, shp.text):
            itemDict = {
                "type": "txt",
                "st": (item.mapToScene(item.start) - item.scene().origin).toTuple(),
                "tc": item.textContent,
                "ff": item.fontFamily,
                "fs": item.fontStyle,
                "th": item.textHeight,
                "ta": item.textAlignment,
                "to": item.textOrient,
                "ang": item.angle,
            }
            return itemDict
        # elif isinstance(item, net.crossingDot):
        #     itemDict = {
        #         "type": "dot",
        #         "pt": (item.mapToScene(item.point) - item.scene().origin).toTuple(),
        #     }
        #     return itemDict



import sys
from PySide6.QtCore import QRunnable, Slot, Qt
from PySide6.QtWidgets import (
    QApplication,
    QMainWindow,
    QVBoxLayout,
    QWidget,
    QDialog,
    QLabel,
    QPushButton,
)

from PySide6.QtGui import QAction, QIcon
from PySide6.QtWebEngineWidgets import (
    QWebEngineView,
)
from PySide6.QtPrintSupport import (
    QPrintDialog,
    QPrinter,
    QPrintPreviewDialog,
)


class helpBrowser(QMainWindow):
    def __init__(self, parent):
        super().__init__(parent=parent)

        self.initUI()

    def initUI(self):
        """
        Initializes the UI of the Help Browser window.
        """
        # Set window title
        self.setWindowTitle("Help Browser")

        # Set window geometry
        self.setGeometry(100, 100, 800, 600)

        # Create central widget
        self.centralW = QWidget(self)

        # Create layout for central widget
        layout = QVBoxLayout(self.centralW)

        # Create web view
        webView = QWebEngineView()

        # Add web view to layout
        layout.addWidget(webView)

        # Set central widget
        self.setCentralWidget(self.centralW)

        # Create menu bar
        menuBar = self.menuBar()

        # Create 'Print' menu
        printMenu = menuBar.addMenu("Print")

        # Create print action
        printIcon = QIcon(":/icons/printer--arrow.png")
        self.printAction = QAction(printIcon, "Print...", self)

        # Create print preview action
        printPreviewIcon = QIcon(":/icons/printer--arrow.png")
        self.printPreviewAction = QAction(printPreviewIcon, "Print Preview...", self)

        # Connect print action to printClick method
        self.printAction.triggered.connect(self.printClick)

        # Connect print preview action to printPreviewClick method
        self.printPreviewAction.triggered.connect(self.printPreviewClick)

        # Add print and print preview actions to print menu
        printMenu.addAction(self.printAction)
        printMenu.addAction(self.printPreviewAction)

        # Set website URL for web view
        websiteUrl = "https://www.reveda.eu/documentation/"
        self.centralW.layout().itemAt(0).widget().setUrl(websiteUrl)

    def printClick(self):
        """
        Handles the 'Print' action.
        """
        # Create print dialog
        dlg = QPrintDialog(self)

        # If print dialog is accepted, start printing
        if dlg.exec() == QDialog.Accepted:
            # Get printer
            printer = dlg.printer()

            # Start print view in a separate thread
            printRunner = startThread(self.printView(printer))
            self.appMainW.threadPool.start(printRunner)

            # Log printing started
            self.logger.info("Printing started")

    def printPreviewClick(self):
        """
        Handles the 'Print Preview' action.
        """
        # Create printer with screen resolution
        printer = QPrinter(QPrinter.ScreenResolution)

        # Set output format to PDF
        printer.setOutputFormat(QPrinter.PdfFormat)

        # Create print preview dialog
        ppdlg = QPrintPreviewDialog(self)

        # Connect paintRequested signal to printView method
        ppdlg.paintRequested.connect(self.printView)

        # Show print preview dialog
        ppdlg.exec()


class aboutDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)

        self.setWindowTitle("About Revolution EDA")
        self.setWindowFlags(Qt.Dialog | Qt.WindowStaysOnTopHint)
        self.setMinimumWidth(300)

        layout = QVBoxLayout()

        # Add information about your application using rich text
        about_label = QLabel(
            "<h2>Revolution EDA</h2>"
            "<p><strong>Version:</strong> 0.6</p>"
            "<p><strong>Copyright: Revolution Semiconductor</strong> © 2024</p>"
            "<p><strong>License:</strong> Mozilla Public License 2.0 amended with Commons Clause</p>"
            "<p><strong> Website:</strong> <a href='https://www.reveda.eu'>Revolution EDA</a></p>"
            "<p><strong> GitHub:</strong> <a href='https://github.com/eskiyerli/revedaRelease'>Revolution EDA GitHub Repository</a></p>"
            "<p><strong> Inquiries:</strong> <a href='mailto:info@revsemi.com'>info@revsemi.com</a></p>"
        )
        about_label.setOpenExternalLinks(True)  # Allow clickable links
        layout.addWidget(about_label)
        layout.addSpacing(20)

        # Add a "Close" button
        close_button = QPushButton("Close")
        close_button.clicked.connect(self.accept)
        layout.addWidget(close_button)

        self.setLayout(layout)


def show_about_dialog():
    dialog = AboutDialog()
    dialog.exec_()


if __name__ == "__main__":
    app = QApplication(sys.argv)

    main_window = QMainWindow()
    main_window.setWindowTitle("Revolution EDA")
    main_window.setGeometry(100, 100, 800, 600)

    # Create an "About" action
    about_action = QAction("About", main_window)
    about_action.setShortcut(QKeySequence(Qt.Key_F1))
    about_action.triggered.connect(show_about_dialog)

    # Create a menu bar
    menu_bar = main_window.menuBar()
    help_menu = menu_bar.addMenu("Help")
    help_menu.addAction(about_action)

    main_window.show()
    sys.exit(app.exec_())


class startThread(QRunnable):
    __slots__ = ("fn",)

    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    @Slot()
    def run(self) -> None:
        try:
            self.fn
        except Exception as e:
            print(e)


def main():
    app = QApplication(sys.argv)
    window = helpBrowser()

    window.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()
